#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import re
from distutils import log as logger
from distutils.dir_util import remove_tree

import pip

try:
    from setuptools import Command
except ImportError:
    from distutils.core import Command

from packaging import package_dir


class bdist_trinoadmin(Command):

    description = 'create a distribution for trino-admin'

    user_options = [('bdist-dir=', 'b',
                     'temporary directory for creating the distribution'),
                    ('dist-dir=', 'd',
                     'directory to put final built distributions in'),
                    ('virtualenv-version=', None,
                     'version of virtualenv to download'),
                    ('keep-temp', 'k',
                     'keep the pseudo-installation tree around after ' +
                     'creating the distribution archive'),
                    ('online-install', None, 'boolean flag indicating if ' +
                     'the installation should pull dependencies from the ' +
                     'Internet or use the ones supplied in the third party ' +
                     'directory'),
                    ('plat-name=', 'p',
                     'platform name to embed in generated filenames' +
                     '(default: linux_x86_64)')
                    ]

    default_virtualenv_version = '12.0.7'

    NATIVE_WHEELS = ['pycrypto-2.6.1-{0}-none-linux_x86_64.whl', 'twofish-0.3.0-{0}-none-linux_x86_64.whl']

    def build_wheel(self, build_dir):
        cmd = self.reinitialize_command('bdist_wheel')
        cmd.dist_dir = build_dir
        self.run_command('bdist_wheel')
        cmd.compression = 'deflated'
        # Ensure that you get the finalized archive name
        cmd.finalize_options()
        # wheel_name = cmd.get_archive_basename()
        # logger.info('creating %s in %s', wheel_name + '.whl', build_dir)

        return ""

    def generate_install_script(self, wheel_name, build_dir):
        with open(os.path.join(package_dir, 'install-trinoadmin.template'), 'r') as template:
            with open(os.path.join(build_dir, 'install-trinoadmin.sh'), 'w') as install_script_file:
                install_script = self._fill_in_template(template.readlines(), wheel_name)
                install_script_file.write(install_script)
                os.chmod(os.path.join(build_dir, 'install-trinoadmin.sh'), 0o755)

    def _fill_in_template(self, template_lines, wheel_name):
        if self.online_install:
            extra_install_args = ''
        else:
            extra_install_args = '--no-index --find-links third-party'

        filled_in = [self._replace_template_values(line, wheel_name, extra_install_args) for line in template_lines]
        return ''.join(filled_in)

    def _replace_template_values(self, line, wheel_name, extra_install_args):
        line = re.sub(r'%ONLINE_OR_OFFLINE_INSTALL%', extra_install_args, line)
        line = re.sub(r'%WHEEL_NAME%', wheel_name, line)
        line = re.sub(r'%VIRTUALENV_VERSION%', self.virtualenv_version, line)
        return line

    def package_dependencies(self, build_dir):
        thirdparty_dir = os.path.join(build_dir, 'third-party')

        requirements = self.distribution.install_requires
        for requirement in requirements:
            pip.main(['wheel',
                      '--wheel-dir={0}'.format(thirdparty_dir),
                      '--no-cache',
                      requirement])

        pip.main(['download',
                  '-d',
                  thirdparty_dir,
                  '--no-cache-dir',
                  '--no-binary',
                  ':all:',
                  'virtualenv=={0}'.format(self.virtualenv_version)])

    def archive_dist(self, build_dir, dist_dir):
        archive_basename = self.distribution.get_fullname()
        if self.online_install:
            archive_basename += '-online'
        else:
            archive_basename += '-offline'
        archive_file = os.path.join(dist_dir, archive_basename)
        self.mkpath(os.path.dirname(archive_file))
        self.make_archive(archive_file, 'gztar',
                          root_dir=os.path.dirname(build_dir),
                          base_dir=os.path.basename(build_dir))
        logger.info('created %s.tar.gz', archive_file)

    def run(self):
        build_dir = self.bdist_dir
        self.mkpath(build_dir)

        wheel_name = self.build_wheel(build_dir)
        self.generate_install_script(wheel_name, build_dir)
        if not self.online_install:
            self.package_dependencies(build_dir)

        self.archive_dist(build_dir, self.dist_dir)

        if not self.keep_temp:
            remove_tree(build_dir)

    def initialize_options(self):
        self.bdist_dir = None
        self.dist_dir = None
        self.virtualenv_url_base = None
        self.virtualenv_version = None
        self.keep_temp = False
        self.online_install = False
        self.plat_name = None

    def finalize_options(self):
        if self.bdist_dir is None:
            bdist_base = self.get_finalized_command('bdist').bdist_base
            self.bdist_dir = os.path.join(bdist_base, self.distribution.get_name())

        if self.dist_dir is None:
            self.dist_dir = 'dist'

        if self.virtualenv_version is None:
            self.virtualenv_version = self.default_virtualenv_version

        self.plat_name_supplied = self.plat_name is not None




# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for installing, monitoring, and controlling presto server
using presto-admin
"""
import cgi
import logging
import re
import socket
import sys
import urllib2
import urlparse
from contextlib import closing

from fabric.api import task, sudo, env
from fabric.context_managers import settings, hide
from fabric.decorators import runs_once, with_settings
from fabric.operations import run, os
from fabric.tasks import execute
from fabric.utils import warn, error, abort
from retrying import retry, RetryError

import util.filesystem
from trinoadmin import catalog
from trinoadmin import configure_cmds
from trinoadmin import package
from trinoadmin.standalone.config import StandaloneConfig
from trinoadmin.trinoclient import TrinoClient
from trinoadmin.util import constants
from trinoadmin.util.base_config import requires_config
from trinoadmin.util.exception import ConfigFileNotFoundError, ConfigurationError
from trinoadmin.util.fabricapi import get_host_list, get_coordinator_role
from trinoadmin.util.local_config_util import get_catalog_directory
from trinoadmin.util.remote_config_util import lookup_port, \
    lookup_server_log_file, lookup_launcher_log_file, lookup_string_config

__all__ = ['install', 'uninstall', 'upgrade', 'start', 'stop', 'restart',
           'status']

if constants.BRAND == 'trino':
    INIT_SCRIPTS = '/etc/init.d/trino'
else:
    INIT_SCRIPTS = '/etc/init.d/presto'

RETRY_TIMEOUT = 120
SYSTEM_RUNTIME_NODES = 'select * from system.runtime.nodes'


def old_sysnode_processor(node_info_rows):
    def old_transform(node_is_active):
        return 'active' if node_is_active else 'inactive'

    return get_sysnode_info_from(node_info_rows, old_transform)


def new_sysnode_processor(node_info_rows):
    return get_sysnode_info_from(node_info_rows, lambda x: x)


# NODE_INFO_PER_URI_SQL = VersionRangeList(
#     VersionRange((0, 0), (0, 128),
#                  ('select http_uri, node_version, active from '
#                   'system.runtime.nodes where '
#                   'url_extract_host(http_uri) = \'%s\'',
#                   old_sysnode_processor)),
#     VersionRange((0, 128), (sys.maxsize,),
#                  ('select http_uri, node_version, state from '
#                   'system.runtime.nodes where '
#                   'url_extract_host(http_uri) = \'%s\'',
#                   new_sysnode_processor))
# )
NODE_INFO_PER_URI_SQL = 'select http_uri, node_version, state from ' + \
                        'system.runtime.nodes where ' + \
                        'url_extract_host(http_uri) = \'%s\''

EXTERNAL_IP_SQL = 'select url_extract_host(http_uri) from ' \
                  'system.runtime.nodes WHERE node_id = \'%s\''
CATALOG_INFO_SQL = 'select catalog_name from system.metadata.catalogs'
_LOGGER = logging.getLogger(__name__)

DOWNLOAD_DIRECTORY = '/tmp'
DEFAULT_RPM_NAME = constants.BRAND + '-server-rpm.rpm'
if constants.BRAND == 'trino':
    LATEST_RPM_URL = 'https://repository.sonatype.org/service/local/artifact/maven' \
                     '/content?r=central-proxy&g=io.trinodb' \
                     '&a=trino-server-rpm&e=rpm&v=RELEASE'
else:
    LATEST_RPM_URL = 'https://repository.sonatype.org/service/local/artifact/maven' \
                     '/content?r=central-proxy&g=com.facebook.presto' \
                     '&a=presto-server-rpm&e=rpm&v=RELEASE'


class LocalPrestoRpmFinder:
    def __init__(self, local_path):
        self.local_path = local_path

    @staticmethod
    def _check_rpm_uncorrupted(rpm_path):
        # package.check_if_valid_rpm() outputs information that is not applicable
        # to this function
        # stderr is redirected to not be displayed and should be restored at the
        # end of the function to behave as expected later
        old_stderr = sys.stderr
        sys.stderr = open(os.devnull, 'w')
        try:
            package.check_if_valid_rpm(rpm_path)
        except SystemExit:
            try:
                os.remove(rpm_path)
                warn('Removed corrupted rpm at: %s' % rpm_path)
            except OSError:
                pass
            return False
        finally:
            sys.stderr = old_stderr

        return True

    def _check_if_absolute_path(self):
        if os.path.isfile(self.local_path) and self._check_rpm_uncorrupted(self.local_path):
            print('Found existing rpm at: %s' % self.local_path)
            return self.local_path
        else:
            return None

    def _check_if_relative_path(self, directory_path):
        path_relative_to_download_dir = os.path.join(directory_path, self.local_path)
        if os.path.isfile(path_relative_to_download_dir) and self._check_rpm_uncorrupted(path_relative_to_download_dir):
            print('Found existing rpm at: %s' % path_relative_to_download_dir)
            return path_relative_to_download_dir
        else:
            return None

    def find_local_presto_rpm(self):
        rpm_at_absolute_path = self._check_if_absolute_path()
        if rpm_at_absolute_path:
            return rpm_at_absolute_path

        rpm_at_relative_path = self._check_if_relative_path(DOWNLOAD_DIRECTORY)
        if rpm_at_relative_path:
            return rpm_at_relative_path

        return None


class UrlHandler:
    def __init__(self, url):
        self.url = url
        self.url_response = None
        try:
            self.url_response = urllib2.urlopen(self.url)
        except urllib2.HTTPError as e:
            _LOGGER.error('Url %s responded with code %s' % (url, e.code))
            raise

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close_url()

    def get_url(self):
        return self.url_response.geturl()

    def get_content_length(self):
        try:
            headers = self.url_response.info()
            return int(headers['Content-Length'])
        except (KeyError, ValueError):
            # Handle the case when the server does not include
            # the 'Content-Length' header
            return None

    def get_download_file_name(self, version=None):
        try:
            headers = self.url_response.info()
            content_disposition = headers['Content-Disposition']
            values, params = cgi.parse_header(content_disposition)
            return params['filename']
        except KeyError:
            # Handle the case when the server does not include
            # the 'Content-Disposition' header
            if not version:
                return DEFAULT_RPM_NAME
            else:
                return DEFAULT_RPM_NAME.rsplit('.', 1)[0] + version + '.rpm'

    def read_block(self, block_size):
        return self.url_response.read(block_size)

    def close_url(self):
        if self.url_response:
            self.url_response.close()


class PrestoRpmDownloader:
    def __init__(self, url_handler):
        self.url_handler = url_handler

    def download_rpm(self, version=None):
        content_length = self.url_handler.get_content_length()
        download_file_path = self.get_download_file_path(version)

        with open(download_file_path, 'wb') as local_file:
            bytes_read = 0
            block_size = 16 * 1024 * 1024
            while True:
                download_buffer = self.url_handler.read_block(block_size)
                if not download_buffer:
                    break
                bytes_read += len(download_buffer)
                local_file.write(download_buffer)
                self.print_download_status(bytes_read, content_length)
            print("Downloaded %d bytes" % bytes_read)

        print('Rpm downloaded to: %s' % download_file_path)
        return download_file_path

    def get_download_file_path(self, version=None):
        return os.path.join(DOWNLOAD_DIRECTORY, self.url_handler.get_download_file_name(version))

    @staticmethod
    def print_download_status(bytes_read, content_length):
        if content_length:
            percent = float(bytes_read) / content_length
            percent = round(percent * 100, 2)
            print('Downloaded %d of %d bytes (%0.2f%%)' %
                  (bytes_read, content_length, percent))


class PrestoRpmFetcher:
    def __init__(self, rpm_specifier):
        self.rpm_specifier = rpm_specifier

    def check_valid_version(self):
        return re.match('^[0-9]+(\.[0-9]+){0,2}$', self.rpm_specifier)

    @staticmethod
    def _find_or_download_latest_presto_rpm():
        return PrestoRpmFetcher.find_or_download_rpm_by_url(LATEST_RPM_URL)

    def use_rpm_specifier_as_latest(self):
        print('Using rpm_specifier as "latest"\n'
              'Fetching the latest presto rpm')
        return self._find_or_download_latest_presto_rpm()

    def _find_or_download_rpm_by_version(self, rpm_version):
        # See here for more information: http://search.maven.org/#api
        if constants.BRAND == 'trino':
            download_url = 'http://search.maven.org/remotecontent?filepath=io/trinodb/' \
                           'trino-server-rpm/' + rpm_version + '/trino-server-rpm-' + \
                           rpm_version + '.rpm'
        else:
            download_url = 'http://search.maven.org/remotecontent?filepath=com/facebook/presto/' \
                           'presto-server-rpm/' + rpm_version + '/presto-server-rpm-' + \
                           rpm_version + '.rpm'

        return self.find_or_download_rpm_by_url(download_url, rpm_version)

    def use_rpm_specifier_as_version(self):
        print('Using rpm_specifier as a version\n'
              'Fetching presto rpm version %s' % self.rpm_specifier)
        return self._find_or_download_rpm_by_version(self.rpm_specifier)

    def use_rpm_specifier_as_url(self):
        print('Using rpm_specifier as a url\n'
              'Fetching presto rpm at url: %s' % self.rpm_specifier)
        return self.find_or_download_rpm_by_url(self.rpm_specifier)

    def use_rpm_specifier_as_local_path(self):
        print('Using rpm_specifier as a local path\n'
              'Fetching local presto rpm at path: %s' % self.rpm_specifier)
        local_finder = LocalPrestoRpmFinder(self.rpm_specifier)
        return local_finder.find_local_presto_rpm()

    @staticmethod
    def find_or_download_rpm_by_url(url, version=None):
        """
        Args:
            url:      The url of the presto rpm to be downloaded.
            version:  An optional version number.
                      If the server doesn't respond with the file name that is being
                      requested, this allows the downloaded file to have the correct
                      version attached to its name (presto-server-rpm-'version'.rpm)
                      rather than the default name

        If downloading the presto rpm at the given url would overwrite an existing rpm,
        this function returns the path to the existing rpm. However, if the rpm that
        would be downloaded takes the default rpm name, it will overwrite the existing
        rpm because there is no way to know if the default rpm name is of the same version
        as the requested rpm. If the rpm is corrupted, this function will remove the corrupted
        rpm and attempt to download it.

        Returns:
            The path to the downloaded or found presto rpm
        """
        with UrlHandler(url) as url_handler:
            downloader = PrestoRpmDownloader(url_handler)
            download_file_path = downloader.get_download_file_path(version)
            local_finder = LocalPrestoRpmFinder(download_file_path)
            local_rpm_path = local_finder.find_local_presto_rpm()
            if local_rpm_path and os.path.basename(local_rpm_path) != DEFAULT_RPM_NAME:
                print('Found and using local presto rpm at path: %s\n'
                      'Delete the existing rpm to force a new download' % local_rpm_path)
                return local_rpm_path
            elif local_rpm_path:
                print('Found local presto rpm at path: %s\n'
                      'The rpm has the default name, so it will not be used' % local_rpm_path)
            print('Downloading rpm from %s\n'
                  'to %s\n'
                  'This can take a few minutes' % (url_handler.get_url(), download_file_path))
            return downloader.download_rpm(version)

    def get_path_to_presto_rpm(self):
        """
        This function finds and downloads (if necessary) the requested presto rpm, which can be
        figured out from the rpm_specifier. The rpm_specifier can take many forms:
        'latest', url, version, and local path (from highest to lowest precedence). This function
        interprets the rpm_specifier only as the highest precedence form.
        """
        scheme, netloc, path, parameters, query, fragment = urlparse.urlparse(self.rpm_specifier)
        if self.rpm_specifier == "latest":
            rpm_path = self.use_rpm_specifier_as_latest()
        elif scheme != '' and scheme != 'file':
            rpm_path = self.use_rpm_specifier_as_url()
        elif self.check_valid_version():
            rpm_path = self.use_rpm_specifier_as_version()
        else:
            rpm_path = self.use_rpm_specifier_as_local_path()

        if not rpm_path:
            abort('Unable to find or download presto rpm with specifier %s' % self.rpm_specifier)
        else:
            return rpm_path


@task
@runs_once
@requires_config(StandaloneConfig)
def install(rpm_specifier):
    """
    Copy and install the presto-server rpm to all the nodes in the cluster and
    configure the nodes.

    The topology information will be read from the config.json file. If this
    file is missing, then the coordinator and workers will be obtained
    interactively. Install will fail for invalid json configuration.

    The catalog configurations will be read from the local catalog directory
    which defaults to ~/.prestoadmin/catalog. If this directory is missing or empty
    then no catalog configuration is deployed.

    Install will fail for incorrectly formatted configuration files. Expected
    format is key=value for .properties files and one option per line for
    jvm.config

    Parameters:
        rpm_specifier - String specifying location of presto rpm to copy and install
                        to nodes in the cluster. The string can specify a presto rpm
                        in the following ways:

                        1.  'latest' to download the latest release
                        2.  Url to download
                        3.  Version number to download
                        4.  Path to a local copy

                        If rpm_specifier matches multiple forms, it is interpreted as the form with
                        highest precedence. The forms are listed from highest to lowest precedence
                        (going top to bottom) For example, if the rpm_specifier matches the criteria
                        to be a url to download, it will be interpreted as such and will never be
                        interpreted as a version number or a local path.

                        Before downloading an rpm, install will attempt to find a local
                        copy with a matching version number to the requested rpm. If such
                        a match is found, it will use the local copy instead of downloading
                        the rpm again.

        --nodeps -      (optional) Flag to indicate if server install
                        should ignore checking Presto rpm package
                        dependencies. Equivalent to adding --nodeps
                        flag to rpm -i.

        --no-config-update - pass this in order to avoid server update after installation
    """
    rpm_fetcher = PrestoRpmFetcher(rpm_specifier)
    path_to_rpm = rpm_fetcher.get_path_to_presto_rpm()
    package.check_if_valid_rpm(path_to_rpm)
    return execute(deploy_install_configure, path_to_rpm, hosts=get_host_list())


def _do_config():
    return False if env.no_config_update else True


def deploy_install_configure(local_path):
    package.deploy_install(local_path)
    if _do_config():
        update_configs()
    wait_for_presto_user()


def add_tpch_catalog():
    tpch_catalog_config = os.path.join(get_catalog_directory(), 'tpch.properties')
    util.filesystem.write_to_file_if_not_exists('connector.name=tpch', tpch_catalog_config)


def update_configs():
    configure_cmds.deploy()

    add_tpch_catalog()
    try:
        catalog.add()
    except ConfigFileNotFoundError:
        _LOGGER.info('No catalog directory found, not adding catalogs.')


@retry(stop_max_delay=3000, wait_fixed=250)
def wait_for_presto_user():
    ret = sudo('getent passwd presto', quiet=True)
    if not ret.succeeded:
        raise Exception('Presto package was not installed successfully. '
                        'Presto user was not created.')


@task
@requires_config(StandaloneConfig)
def uninstall():
    """
    Uninstall Presto after stopping the services on all nodes

    Parameters:
        --nodeps -              (optional) Flag to indicate if server uninstall
                                should ignore checking Presto rpm package
                                dependencies. Equivalent to adding --nodeps
                                flag to rpm -e.
    """
    stop()

    if package.is_rpm_installed(constants.BRAND):
        package.rpm_uninstall(constants.BRAND)
    elif package.is_rpm_installed(constants.BRAND + '-server'):
        package.rpm_uninstall(constants.BRAND + '-server')
    elif package.is_rpm_installed(constants.BRAND + '-server-rpm'):
        package.rpm_uninstall(constants.BRAND + '-server-rpm')
    else:
        abort('Unable to uninstall package on: ' + env.host)


@task
@requires_config(StandaloneConfig)
def upgrade(new_rpm_path, local_config_dir=None, overwrite=False):
    """
    Copy and upgrade a new presto-server rpm to all of the nodes in the
    cluster. Retains existing node configuration.

    The existing topology information is read from the config.json file.
    Unlike install, there is no provision to supply topology information
    interactively.

    The existing cluster configuration is collected from the nodes on the
    cluster and stored on the host running presto-admin. After the
    presto-server packages have been upgraded, presto-admin pushes the
    collected configuration back out to the hosts on the cluster.

    Note that the configuration files in the presto-admin configuration
    directory are not updated during upgrade.

    :param new_rpm_path -       The path to the new Presto RPM to
                                install
    :param local_config_dir -   (optional) Directory to store the cluster
                                configuration in. If not specified, a temp
                                directory is used.
    :param overwrite -          (optional) if set to True then existing
                                configuration will be orerwriten.

    :param --nodeps -           (optional) Flag to indicate if server upgrade
                                should ignore checking Presto rpm package
                                dependencies. Equivalent to adding --nodeps
                                flag to rpm -U.
    """
    stop()

    temp_config_tar = configure_cmds.gather_config_directory()

    package.deploy_upgrade(new_rpm_path)

    configure_cmds.deploy_config_directory(temp_config_tar)


def service(control=None):
    if check_presto_version() != '':
        return False
    if control == 'start' and is_port_in_use(env.host):
        return False
    _LOGGER.info('Executing %s on presto server' % control)
    ret = sudo('set -m; ' + INIT_SCRIPTS + ' ' + control)
    return ret.succeeded


def check_status_for_control_commands():
    print('Waiting to make sure we can connect to the Presto server on %s, '
          'please wait. This check will time out after %d minutes if the '
          'server does not respond.'
          % (env.host, (RETRY_TIMEOUT / 60)))
    if check_server_status():
        print('Server started successfully on: ' + env.host)
    else:
        warn('Could not verify server status for: ' + env.host +
             '\nThis could mean that the server failed to start or that there was no coordinator or worker up. '
             'Please check ' + lookup_server_log_file(env.host) + ' and ' +
             lookup_launcher_log_file(env.host))


def is_port_in_use(host):
    _LOGGER.info("Checking if port used by Prestoserver is already in use..")
    try:
        portnum = lookup_port(host)
    except Exception:
        _LOGGER.info("Cannot find port from config.properties. "
                     "Skipping check for port already being used")
        return 0
    with settings(hide('warnings', 'stdout'), warn_only=True):
        output = run('netstat -lnt |grep -E "\<%s\>" |grep LISTEN' % str(portnum))
    if output:
        _LOGGER.info("Presto server port already in use. Skipping "
                     "server start...")
        error('Server failed to start on %s. Port %s already in use'
              % (env.host, str(portnum)))
    return output


@task
@requires_config(StandaloneConfig)
def start():
    """
    Start the Presto server on all nodes

    A status check is performed on the entire cluster and a list of
    servers that did not start, if any, are reported at the end.
    """
    if service('start'):
        check_status_for_control_commands()


@task
@requires_config(StandaloneConfig)
def stop():
    """
    Stop the Presto server on all nodes
    """
    service('stop')


def stop_and_start():
    if check_presto_version() != '':
        return False
    sudo('set -m; ' + INIT_SCRIPTS + ' stop')
    if is_port_in_use(env.host):
        return False
    _LOGGER.info('Executing start on presto server')
    ret = sudo('set -m; ' + INIT_SCRIPTS + ' start')
    return ret.succeeded


@task
@requires_config(StandaloneConfig)
def restart():
    """
    Restart the Presto server on all nodes.

    A status check is performed on the entire cluster and a list of
    servers that did not start, if any, are reported at the end.
    """
    if stop_and_start():
        check_status_for_control_commands()


def check_presto_version():
    """
    Checks that the Presto version is suitable.

    Returns:
        Error string if applicable
    """
    if not presto_installed():
        not_installed_str = constants.BRAND.capitalize() + ' is not installed.'
        warn(not_installed_str)
        return not_installed_str

    return ''


def presto_installed():
    with settings(hide('warnings', 'stdout'), warn_only=True):
        package_search = run('rpm -q ' + constants.BRAND + '-server-rpm')
        if not package_search.succeeded:
            package_search = run('rpm -q ' + constants.BRAND)
        if not package_search.succeeded:
            package_search = run('rpm -q starburst-presto-server-rpm')
        return package_search.succeeded


def get_presto_version():
    with settings(hide('warnings', 'stdout'), warn_only=True):
        version = run('rpm -q --qf \"%{VERSION}\\n\" ' + constants.BRAND + '-server-rpm')
        # currently we have two rpm names out so we need this retry
        if not version.succeeded:
            version = run('rpm -q --qf \"%{VERSION}\\n\" ' + constants.BRAND)
        version = version.strip()
        _LOGGER.debug('{} rpm version: {}'.format(constants.BRAND.capitalize(), version))

        return version


def check_server_status():
    """
    Checks if server is running for env.host. Retries connecting to server
    until server is up or till RETRY_TIMEOUT is reached

    Parameters:
        client - client that executes the query

    Returns:
        True or False
    """
    if len(get_coordinator_role()) < 1:
        warn('No coordinator defined.  Cannot verify server status.')
    with closing(TrinoClient(get_coordinator_role()[0], env.user)) as client:
        node_id = lookup_string_config('node.id', os.path.join(constants.REMOTE_CONF_DIR, 'node.properties'), env.host)
        try:
            return query_server_for_status(client, node_id)
        except RetryError:
            return False


@retry(stop_max_delay=RETRY_TIMEOUT * 1000, wait_fixed=5000, retry_on_result=lambda result: result is False)
def query_server_for_status(client, node_id):
    try:
        rows = client.run_sql(SYSTEM_RUNTIME_NODES)
        if rows is not None:
            return _is_in_rows(node_id, rows)
    except ConfigurationError as e:
        _LOGGER.warn(e)
    return False


def _is_in_rows(value, rows):
    for row in rows:
        if value in row:
            return True
    return False


def execute_catalog_info_sql(client):
    """
    Returns [[catalog_name], [catalog_2]..] from catalogs system table

    Parameters:
        client - client that executes the query
    """
    return client.run_sql(CATALOG_INFO_SQL)


def execute_external_ip_sql(client, uuid):
    """
    Returns external ip of the host with uuid after parsing the http_uri column
    from nodes system table

    Parameters:
        client - client that executes the query
        uuid - node_id of the node
    """
    return client.run_sql(EXTERNAL_IP_SQL % uuid)


def get_sysnode_info_from(node_info_row, state_transform):
    """
    Returns system node info dict from node info row for a node

    Parameters:
        node_info_row -

    Returns:
        Node info dict in format:
        {'http://node1/statement': [presto-main:0.97-SNAPSHOT, True]}
    """
    output = {}
    for row in node_info_row:
        if row:
            output[row[0]] = [row[1], state_transform(row[2])]

    _LOGGER.info('Node info: %s ', output)
    return output


def get_catalog_info_from(client):
    """
    Returns installed catalogs

    Parameters:
        client - client that executes the query

    Returns:
        comma delimited catalogs eg: tpch, hive, system
    """
    syscatalog = []
    catalog_info = execute_catalog_info_sql(client)
    for conn_info in catalog_info:
        if conn_info:
            syscatalog.append(conn_info[0])
    return ', '.join(syscatalog)


def get_roles_for(host):
    roles = []
    for role in ['coordinator', 'worker']:
        if host in env.roledefs[role]:
            roles.append(role)
    return roles


def print_node_info(node_status, catalog_status):
    """

    Args:
        node_status: list, [node_id, http_uri, node_version, coordinator, state]
        catalog_status: list of catalogies

    Returns:

    """

    print('\tNode URI(http): ' + node_status[1] +
          '\n\t' + constants.BRAND.capitalize() + ' Version: ' + node_status[2] +
          '\n\tNode status:    ' + node_status[4])
    if catalog_status:
        print('\tCatalogs:     ' + catalog_status)
    # for k in node_status:
    #     print('\tNode URI(http): ' + str(k) +
    #           '\n\t' + constants.BRAND.capitalize() + ' Version: ' + str(node_status[k][0]) +
    #           '\n\tNode status:    ' + str(node_status[k][1]))
    #     if catalog_status:
    #         print('\tCatalogs:     ' + catalog_status)


def print_status_header(external_ip, server_status, host):
    print('Server Status:')
    print('\t{}(IP: {}, Roles: {}): {}'.format(host, external_ip, ', '.join(get_roles_for(host)), server_status))


def get_status_from_coordinator():
    """
    get all nodes' status according to query "select * from system.runtime.nodes"

    Returns: list of node status, each element is a list, like
        1b6055c4-7e03-4c27-a809-a8284b3fd0e9,http://10.1.1.1:8080,356,true,active
        the first field is node_id, followed by http_uri, node_version, coordinator, state
    """
    _LOGGER.info("Begin to get status from coordinator")
    with closing(TrinoClient(get_coordinator_role()[0], env.user)) as client:
        try:
            coordinator_status = client.run_sql(SYSTEM_RUNTIME_NODES)
            catalog_status = get_catalog_info_from(client)
        except Exception as e:
            # Just log errors that come from a missing port or anything else; if
            # we can't connect to the coordinator, we just want to print out a
            # minimal status anyway.
            _LOGGER.warn("Failed to get coordinator status: " + e.message)
            coordinator_status = []
            catalog_status = []

        for host in get_host_list():
            try:
                external_ip = socket.gethostbyname(host)
                if get_node_status(coordinator_status, external_ip):
                    is_running = "Running"
                else:
                    is_running = "Not Running"
                error_message = ""
            except socket.gaierror as e:
                external_ip = 'Unknow'
                is_running = "Not Running"
                error_message = e.message

            print_status_header(external_ip, is_running, host)
            if error_message:
                print('\t' + error_message)
            elif not coordinator_status:
                print('\tNo information available: unable to query coordinator')
            elif not is_running:
                print('\tNo information available')
            else:
                node_status = get_node_status(coordinator_status, external_ip)
                if node_status:
                    print_node_info(node_status, catalog_status)
                else:
                    print('\tNo information available: the coordinator has not yet'
                          ' discovered this node')


def get_node_status(coordinator_status, ip):
    """

    Args:
        coordinator_status: list of node status
        ip: node ip

    Returns: node status

    """
    for row in coordinator_status:
        if ip in row[1]:
            return row
    return None


@task
@runs_once
@requires_config(StandaloneConfig)
@with_settings(hide('warnings'))
def status():
    """
    Print the status of presto in the cluster
    """
    get_status_from_coordinator()



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Modules within util should only use the standard library because setup.py
# may rely on the modules. setup.py typically installs all dependencies.
# If a third party module is used, setup.py may attempt to import it while
# trying to install dependencies and an ImportError will be raised because
# the dependency has not been installed yet.
import os

main_dir = os.path.dirname(os.path.abspath(os.path.dirname(__file__)))

with open(os.path.join(main_dir, 'trinoadmin/_version.py')) as version_file:
    __version__ = version_file.readlines()[-1].split()[-1].strip("\"'")



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for installing and uninstalling slider.
"""

import os

from fabric.api import env, task, abort
from fabric.operations import put, sudo

from trinoadmin.yarn_slider.config import SliderConfig, \
    DIR, SLIDER_MASTER
from trinoadmin.util.base_config import requires_config

from trinoadmin.util.fabricapi import task_by_rolename

__all__ = ['install', 'uninstall']


@task
@requires_config(SliderConfig)
@task_by_rolename(SLIDER_MASTER)
def install(slider_tarball):
    """
    Install slider on the slider master. You must provide a tar file on the
    local machine that contains the slider distribution.

    :param slider_tarball: The gzipped tar file containing the Apache Slider
                           distribution
    """
    deploy_install(slider_tarball)


def deploy_install(slider_tarball):
    slider_dir = env.conf[DIR]
    slider_parent = os.path.dirname(slider_dir)
    slider_file = os.path.join(slider_parent, os.path.basename(slider_tarball))

    sudo('mkdir -p %s' % (slider_dir))

    result = put(slider_tarball, os.path.join(slider_parent, slider_file))
    if result.failed:
        abort('Failed to send slider tarball %s to directory %s on host %s' %
              (slider_tarball, slider_dir, env.host))

    sudo('gunzip -c %s | tar -x -C %s --strip-components=1 && rm -f %s' %
         (slider_file, slider_dir, slider_file))


@task
@requires_config(SliderConfig)
@task_by_rolename(SLIDER_MASTER)
def uninstall():
    """
    Uninstall slider from the slider master.
    """
    sudo('rm -r "%s"' % (env.conf[DIR]))



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for handling presto-admin mode-related functionality.
"""

import os

from fabric.api import abort, task
from fabric.decorators import runs_once

from trinoadmin import config
from trinoadmin.util.exception import ConfigurationError, \
    ConfigFileNotFoundError
from trinoadmin.util.local_config_util import get_config_directory

MODE_CONF_PATH = os.path.join(get_config_directory(), 'mode.json')
MODE_KEY = 'mode'

MODE_SLIDER = 'yarn_slider'
MODE_STANDALONE = 'standalone'

VALID_MODES = [MODE_SLIDER, MODE_STANDALONE]


def _load_mode_config():
    return config.get_conf_from_json_file(MODE_CONF_PATH)


def _store_mode_config(mode_config):
    config.write(config.json_to_string(mode_config), MODE_CONF_PATH)


def get_mode(validate=True):
    mode_config = _load_mode_config()
    mode = mode_config.get(MODE_KEY)

    if validate and mode is None:
        raise ConfigurationError(
            'Required key %s not found in configuration file %s' % (
                MODE_KEY, MODE_CONF_PATH))

    if validate and not validate_mode(mode):
        raise ConfigurationError(
            'Invalid mode %s in configuration file %s. Valid modes are %s' % (
                mode, MODE_CONF_PATH, ' '.join(VALID_MODES)))

    return mode


def validate_mode(mode):
    return mode in VALID_MODES


def for_mode(mode, mode_map):
    if sorted(mode_map.keys()) != sorted(VALID_MODES):
        raise Exception(
            'keys in for_nodes\n%s\ndo not match VALID_MODES\n%s' % (
                mode_map.keys(), VALID_MODES))
    return mode_map[mode]


@task
@runs_once
def select(new_mode):
    """
    Change the mode.
    """
    if not validate_mode(new_mode):
        abort('Invalid mode selection %s. Valid modes are %s' % (
            new_mode, ' '.join(VALID_MODES)))

    mode_config = {}
    try:
        mode_config = _load_mode_config()
    except ConfigFileNotFoundError:
        pass

    mode_config[MODE_KEY] = new_mode
    _store_mode_config(mode_config)


@task
@runs_once
def get():
    """
    Display the current mode.
    """
    mode = None
    try:
        mode = get_mode(validate=False)
        print mode
    except ConfigFileNotFoundError:
        abort("Select a mode using the subcommand 'mode select <mode>'")


@task
@runs_once
def list():
    """
    List the supported modes.
    """
    print ' '.join(VALID_MODES)



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Module for processing presto configuration files
"""
import logging
import os

from trinoadmin.config import get_conf_from_properties_file, \
    get_conf_from_config_file
from trinoadmin.util.exception import ConfigurationError


REQUIRED_FILES = ["node.properties", "jvm.config", "config.properties"]
PRESTO_FILES = ["node.properties", "jvm.config", "config.properties",
                "log.properties"]
_LOGGER = logging.getLogger(__name__)


def get_presto_conf(conf_dir):
    if os.path.isdir(conf_dir):
        file_list = [name for name in os.listdir(conf_dir) if
                     name in PRESTO_FILES]
    else:
        _LOGGER.debug("No directory " + conf_dir)
        file_list = []

    conf = {}
    for filename in file_list:
        ext = os.path.splitext(filename)[1]
        file_path = os.path.join(conf_dir, filename)
        if ext == ".properties":
            conf[filename] = get_conf_from_properties_file(file_path)
        elif ext == ".config":
            conf[filename] = get_conf_from_config_file(file_path)
    return conf


def validate_presto_conf(conf):
    for required in REQUIRED_FILES:
        if required not in conf:
            raise ConfigurationError("Missing configuration for required "
                                     "file: " + required)

    expect_object_msg = "%s must be an object with key-value property pairs"
    if not isinstance(conf["node.properties"], dict):
        raise ConfigurationError(expect_object_msg % "node.properties")

    if not isinstance(conf["jvm.config"], list):
        raise ConfigurationError("jvm.config must contain a json array of jvm "
                                 "arguments ([arg1, arg2, arg3])")

    if not isinstance(conf["config.properties"], dict):
        raise ConfigurationError(expect_object_msg % "config.properties")

    return conf



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module for gathering various debug information for incident reporting
using presto-admin
"""

import logging
import json
import shutil
import tarfile

import requests
from fabric.contrib.files import append
from fabric.context_managers import settings, hide
from fabric.operations import os, get, run
from fabric.tasks import execute
from fabric.api import env, runs_once, task
from fabric.utils import abort, warn

from trinoadmin.trinoclient import TrinoClient
from trinoadmin.server import get_presto_version, get_catalog_info_from
from trinoadmin.util.base_config import requires_config
from trinoadmin.util.filesystem import ensure_directory_exists
from trinoadmin.util.local_config_util import get_log_directory
from trinoadmin.util.remote_config_util import lookup_server_log_file,\
    lookup_launcher_log_file,  lookup_port, lookup_catalog_directory
from trinoadmin.standalone.config import StandaloneConfig
import trinoadmin.util.fabricapi as fabricapi
import trinoadmin


TMP_PRESTO_DEBUG = '/tmp/trino-debug/'
TMP_PRESTO_DEBUG_REMOTE = '/tmp/trino-debug-remote'
OUTPUT_FILENAME_FOR_LOGS = 'trino-debug-logs.tar.gz'
DEFAULT_PATH_FOR_LOGS = TMP_PRESTO_DEBUG + OUTPUT_FILENAME_FOR_LOGS
OUTPUT_FILENAME_FOR_SYS_INFO = '/tmp/trino-debug-sysinfo.tar.gz'
PRESTOADMIN_LOG_NAME = 'trino-admin.log'
_LOGGER = logging.getLogger(__name__)
QUERY_REQUEST_EXT = 'v1/query/'
NODES_REQUEST_EXT = 'v1/node'

__all__ = ['logs', 'query_info', 'system_info']


@task
@runs_once
@requires_config(StandaloneConfig)
def logs(dest_dir=TMP_PRESTO_DEBUG):
    """
    Gather all the server logs and presto-admin log and create a tar file.

    Parameters:
        dest_dir - logs destination dir.  default /tmp/presto-debug
    """
    downloaded_logs_location = os.path.join(dest_dir, "logs")
    ensure_directory_exists(downloaded_logs_location)

    print 'Downloading logs from all the nodes...'
    execute(get_remote_log_files, downloaded_logs_location, roles=env.roles)

    copy_admin_log(downloaded_logs_location)
    logs_tar_file_location = os.path.join(dest_dir, OUTPUT_FILENAME_FOR_LOGS)
    make_tarfile(logs_tar_file_location, downloaded_logs_location)
    print 'logs archive created: ' + logs_tar_file_location


def copy_admin_log(log_folder):
    shutil.copy(os.path.join(get_log_directory(), PRESTOADMIN_LOG_NAME), log_folder)


def make_tarfile(output_filename, source_dir):
    tar = tarfile.open(output_filename, 'w:gz')

    try:
        tar.add(source_dir, arcname=os.path.basename(source_dir))
    finally:
        tar.close()


def get_remote_log_files(dest_path):
    remote_server_log = lookup_server_log_file(env.host)
    _LOGGER.debug('Logs to be archived on host ' + env.host + ': ' + remote_server_log)
    get_files(remote_server_log + '*', dest_path)

    remote_launcher_log = lookup_launcher_log_file(env.host)
    _LOGGER.debug('LOG directory to be archived on host ' + env.host + ': ' + remote_launcher_log)
    get_files(remote_launcher_log + '*', dest_path)


def get_files(remote_path, local_path):
    path_with_host_name = os.path.join(local_path, env.host)

    try:
        os.makedirs(path_with_host_name)
    except OSError:
        if not os.path.isdir(path_with_host_name):
            raise

    _LOGGER.debug('local path used ' + path_with_host_name)

    try:
        get(remote_path, path_with_host_name, use_sudo=True)
    except SystemExit:
        warn('remote path ' + remote_path + ' not found on ' + env.host)


def request_url(url_extension):
    host = env.host
    port = lookup_port(host)
    return 'http://%(host)s:%(port)i/%(url_ext)s' % {'host': host,
                                                     'port': port,
                                                     'url_ext': url_extension}


@task
@requires_config(StandaloneConfig)
def query_info(query_id):
    """
    Gather information about the query identified by the given
    query_id and store that in a JSON file.

    Parameters:
        query_id - id of the query for which info has to be gathered
    """

    if env.host not in fabricapi.get_coordinator_role():
        return

    err_msg = 'Unable to retrieve information. Please check that the ' \
              'query_id is correct, or check that server is up with ' \
              'command: server status'
    req = get_request(request_url(QUERY_REQUEST_EXT + query_id), err_msg)
    query_info_file_name = os.path.join(TMP_PRESTO_DEBUG, 'query_info_' + query_id + '.json')

    try:
        os.makedirs(TMP_PRESTO_DEBUG)
    except OSError:
        if not os.path.isdir(TMP_PRESTO_DEBUG):
            raise

    with open(query_info_file_name, 'w') as out_file:
        out_file.write(json.dumps(req.json(), indent=4))

    print('Gathered query information in file: ' + query_info_file_name)


def get_request(url, err_msg):
        try:
            req = requests.get(url, auth=("'%s'" % (env.user,), ''))
        except requests.ConnectionError:
            abort(err_msg)

        if not req.status_code == requests.codes.ok:
            abort(err_msg)

        return req


@task
@requires_config(StandaloneConfig)
def system_info():
    """
    Gather system information like nodes in the system, presto
    version, presto-admin version, os version etc.
    """
    if env.host not in fabricapi.get_coordinator_role():
        return
    err_msg = 'Unable to access node information. ' \
              'Please check that server is up with command: server status'
    req = get_request(request_url(NODES_REQUEST_EXT), err_msg)

    downloaded_sys_info_loc = os.path.join(TMP_PRESTO_DEBUG, "sysinfo")
    try:
        os.makedirs(downloaded_sys_info_loc)
    except OSError:
        if not os.path.isdir(downloaded_sys_info_loc):
            raise

    node_info_file_name = os.path.join(downloaded_sys_info_loc, 'node_info.json')
    with open(node_info_file_name, 'w') as out_file:
        out_file.write(json.dumps(req.json(), indent=4))

    _LOGGER.debug('Gathered node information in file: ' + node_info_file_name)

    catalog_file_name = os.path.join(downloaded_sys_info_loc, 'catalog_info.txt')
    client = TrinoClient(env.host, env.user)
    catalog_info = get_catalog_info_from(client)

    with open(catalog_file_name, 'w') as out_file:
        out_file.write(catalog_info + '\n')

    _LOGGER.debug('Gathered catalog information in file: ' + catalog_file_name)

    execute(get_catalog_configs, downloaded_sys_info_loc, roles=env.roles)
    execute(get_system_info, downloaded_sys_info_loc, roles=env.roles)

    make_tarfile(OUTPUT_FILENAME_FOR_SYS_INFO, downloaded_sys_info_loc)
    print 'System info archive created: ' + OUTPUT_FILENAME_FOR_SYS_INFO


def get_system_info(download_location):

    run("mkdir -p " + TMP_PRESTO_DEBUG_REMOTE)

    version_file_name = os.path.join(TMP_PRESTO_DEBUG_REMOTE, 'version_info.txt')
    run('rm -f ' + version_file_name)

    append(version_file_name, "platform information : " +
           get_platform_information() + '\n')
    append(version_file_name, 'Java version: ' + get_java_version() + '\n')
    append(version_file_name, 'Presto-admin version: ' +
           trinoadmin.__version__ + '\n')
    append(version_file_name, 'Presto server version: ' +
           get_presto_version() + '\n')

    _LOGGER.debug('Gathered version information in file: ' + version_file_name)

    get_files(version_file_name, download_location)


def get_catalog_configs(dest_path):
    remote_catalog_dir = lookup_catalog_directory(env.host)
    _LOGGER.debug('catalogs to be archived on host ' + env.host + ': ' + remote_catalog_dir)
    get_files(remote_catalog_dir, dest_path)


def get_platform_information():
    with settings(hide('warnings', 'stdout'), warn_only=True):
        platform_info = run('uname -a')
        _LOGGER.debug('platform info: ' + platform_info)
        return platform_info


def get_java_version():
    with settings(hide('warnings', 'stdout'), warn_only=True):
        version = run('java -version')
        _LOGGER.debug('java version: ' + version)
        return version



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Commands for running scripts on a cluster
"""
import logging
from fabric.operations import put, sudo
from fabric.decorators import task
from fabric.api import env
from os import path

from trinoadmin.standalone.config import StandaloneConfig
from trinoadmin.util.base_config import requires_config
from trinoadmin.util.constants import REMOTE_COPY_DIR
from trinoadmin.plugin import write

_LOGGER = logging.getLogger(__name__)
__all__ = ['run', 'copy']


@task
@requires_config(StandaloneConfig)
def run(script, remote_dir='/tmp'):
    """
    Run an arbitrary script on all nodes in the cluster.

    Parameters:
        script - The path to the script
        remote_dir - Where to put the script on the cluster.  Default is /tmp.
    """
    script_name = path.basename(script)
    remote_path = path.join(remote_dir, script_name)
    put(script, remote_path)
    sudo('chmod u+x %s' % remote_path)
    sudo(remote_path)
    sudo('rm %s' % remote_path)


@task
@requires_config(StandaloneConfig)
def copy(local_file, remote_dir=REMOTE_COPY_DIR):
    """
    Copy a file to all nodes in the cluster.

    Parameters:
        local_file - The path to the file
        remote_dir - Where to put the file on the cluster.  Default is /tmp.
    """
    _LOGGER.info('copying file to %s' % env.host)
    write(local_file, remote_dir)



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module for setting and validating the presto-admin Apache Slider config
"""

import os

from overrides import overrides

from fabric.state import env

from trinoadmin.util.base_config import BaseConfig, SingleConfigItem, \
    MultiConfigItem
from trinoadmin.util.local_config_util import get_config_directory
from trinoadmin.util.validators import validate_host, validate_port, \
    validate_username, validate_can_connect, validate_can_sudo

SLIDER_CONFIG_LOADED = 'slider_config_loaded'
SLIDER_CONFIG_DIR = os.path.join(get_config_directory(), 'slider')
SLIDER_CONFIG_PATH = os.path.join(SLIDER_CONFIG_DIR, 'config.json')
SLIDER_MASTER = 'slider_master'

HOST = 'slider_master'
ADMIN_USER = 'admin'
SSH_PORT = 'ssh_port'

DIR = 'slider_directory'
APPNAME = 'slider_appname'
INSTANCE_NAME = 'slider_instname'
SLIDER_USER = 'slider_user'
JAVA_HOME = 'JAVA_HOME'
HADOOP_CONF = 'HADOOP_CONF'

# This key comes from the server install step, NOT a user prompt. Accordingly,
# there is no SliderConfigItem for it in _SLIDER_CONFIG
PRESTO_PACKAGE = 'presto_slider_package'


_SLIDER_CONFIG = [
    MultiConfigItem([
        SingleConfigItem(HOST, 'Enter the hostname for the slider master:',
                         'localhost', validate_host),
        SingleConfigItem(ADMIN_USER, 'Enter the user name to use when ' +
                         'installing slider on the slider master:',
                         'root', validate_username),
        SingleConfigItem(SSH_PORT, 'Enter the port number for SSH ' +
                         'connections to the slider master', 22,
                         validate_port)],
                    validate_can_connect, (ADMIN_USER, HOST, SSH_PORT),
                    'Connection failed for %%(%s)s@%%(%s)s:%%(%s)d. ' +
                    'Re-enter connection information.'),

    SingleConfigItem(DIR, 'Enter the directory to install slider into on '
                     'the slider master:', '/opt/slider', None),

    MultiConfigItem([
        SingleConfigItem(SLIDER_USER, 'Enter a user name for running slider '
                         'on the slider master ', 'yarn',
                         validate_username)],
                    validate_can_sudo,
                    (SLIDER_USER, ADMIN_USER, HOST, SSH_PORT),
                    'Failed to sudo to user %%(%s)s while connecting as ' +
                    '%%(%s)s@%%(%s)s:%%(%s)d. Enter a new username and try' +
                    'again.'),

    SingleConfigItem(JAVA_HOME, 'Enter the value of JAVA_HOME to use when' +
                     'running slider on the slider master:',
                     '/usr/lib/jvm/java', None),
    SingleConfigItem(HADOOP_CONF, 'Enter the location of the Hadoop ' +
                     'configuration on the slider master:',
                     '/etc/hadoop/conf', None),
    SingleConfigItem(APPNAME, 'Enter a name for the presto slider application',
                     'PRESTO', None)]


class SliderConfig(BaseConfig):
    '''
    presto-admin needs to update the slider config other than through the
    interactive config process because it needs to keep track of the name
    of the presto-yarn-package we install.

    As a result, SliderConfig acts a little funny; it acts like enough of a
    dict to allow env.conf[NAME] lookups and modifications, and it also exposes
    the ability to store the config after it's been modified.
    '''

    def __init__(self):
        super(SliderConfig, self).__init__(SLIDER_CONFIG_PATH, _SLIDER_CONFIG)

    @overrides
    def is_config_loaded(self):
        return SLIDER_CONFIG_LOADED in env and env[SLIDER_CONFIG_LOADED]

    @overrides
    def set_config_loaded(self):
        env[SLIDER_CONFIG_LOADED] = True

    @overrides
    def set_env_from_conf(self, conf):
        self.config.update(conf)
        env.user = conf[ADMIN_USER]
        env.port = conf[SSH_PORT]
        env.roledefs[SLIDER_MASTER] = [conf[HOST]]
        env.roledefs['all'] = env.roledefs[SLIDER_MASTER]

        env.conf = self

        env.hosts = env.roledefs['all'][:]

    def store_conf(self):
        super(SliderConfig, self).write_conf(self.config)



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os

from trinoadmin.util.constants import LOG_DIR_ENV_VARIABLE, CONFIG_DIR_ENV_VARIABLE, DEFAULT_LOCAL_CONF_DIR, \
    TOPOLOGY_CONFIG_FILE, COORDINATOR_DIR_NAME, WORKERS_DIR_NAME, CATALOG_DIR_NAME


def get_config_directory():
    config_directory = os.environ.get(CONFIG_DIR_ENV_VARIABLE)
    if not config_directory:
        config_directory = DEFAULT_LOCAL_CONF_DIR
    return config_directory


def get_log_directory():
    config_directory = os.environ.get(LOG_DIR_ENV_VARIABLE)
    if not config_directory:
        config_directory = os.path.join(get_config_directory(), 'log')
    return config_directory


def get_topology_path():
    return os.path.join(get_config_directory(), TOPOLOGY_CONFIG_FILE)


def get_coordinator_directory():
    return os.path.join(get_config_directory(), COORDINATOR_DIR_NAME)


def get_workers_directory():
    return os.path.join(get_config_directory(), WORKERS_DIR_NAME)


def get_catalog_directory():
    return os.path.join(get_config_directory(), CATALOG_DIR_NAME)



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Common module for deploying the presto configuration
"""

import logging
import os

from fabric.contrib import files
from fabric.context_managers import settings
from fabric.contrib.files import exists
from fabric.operations import sudo, abort
from fabric.api import env

from trinoadmin.util import constants
from trinoadmin.standalone.config import PRESTO_STANDALONE_USER_GROUP
import coordinator as coord
import trinoadmin.util.fabricapi as util
import workers as w

_LOGGER = logging.getLogger(__name__)


def coordinator():
    """
    Deploy the coordinator configuration to the coordinator node
    """
    if env.host in util.get_coordinator_role():
        _LOGGER.info("Setting coordinator configuration for " + env.host)
        configure_presto(coord.Coordinator().get_conf(),
                         constants.REMOTE_CONF_DIR)


def workers():
    """
    Deploy workers configuration to the worker nodes.
    This will not deploy configuration for a coordinator that is also a worker
    """
    if env.host in util.get_worker_role() and env.host \
            not in util.get_coordinator_role():
        _LOGGER.info("Setting worker configuration for " + env.host)
        configure_presto(w.Worker().get_conf(), constants.REMOTE_CONF_DIR)


def configure_presto(conf, remote_dir):
    print("Deploying configuration on: " + env.host)
    deploy(dict((name, output_format(content)) for (name, content)
                in conf.iteritems() if name != "node.properties"), remote_dir)
    deploy_node_properties(output_format(conf['node.properties']), remote_dir)


def output_format(conf):
    try:
        return dict_to_equal_format(conf)
    except AttributeError:
        pass
    try:
        return list_to_line_separated(conf)
    except TypeError:
        pass
    except AssertionError:
        pass
    return str(conf)


def dict_to_equal_format(conf):
    sorted_list = sorted(key_val_to_equal(conf.iteritems()))
    return list_to_line_separated(sorted_list)


def key_val_to_equal(items):
    return ["=".join(item) for item in items]


def list_to_line_separated(conf):
    assert not isinstance(conf, basestring)
    return "\n".join(conf)


def deploy(confs, remote_dir):
    _LOGGER.info("Deploying configurations for " + str(confs.keys()))
    sudo("mkdir -p " + remote_dir)
    for name, content in confs.iteritems():
        write_to_remote_file(content, os.path.join(remote_dir, name),
                             owner=PRESTO_STANDALONE_USER_GROUP, mode=600)


def secure_create_file(filepath, user_group, mode=600):
    user, group = user_group.split(':')
    missing_owner_code = 42
    command = \
        "( getent passwd {user} >/dev/null || exit {missing_owner_code} ) &&" \
        " echo '' > {filepath} && " \
        "chown {user_group} {filepath} && " \
        "chmod {mode} {filepath} ".format(
            filepath=filepath, user=user, user_group=user_group, mode=mode,
            missing_owner_code=missing_owner_code)

    with settings(warn_only=True):
        result = sudo(command)
        if result.return_code == missing_owner_code:
            abort("User %s does not exist. Make sure the Presto server RPM "
                  "is installed and try again" % user)
        elif result.failed:
            abort("Failed to securely create file %s" % (filepath))


def secure_create_directory(filepath, user_group, mode=755):
    user, group = user_group.split(':')
    missing_owner_code = 42
    command = \
        "( getent passwd {user} >/dev/null || exit {missing_owner_code} ) && " \
        "mkdir -p {filepath} && " \
        "chown {user_group} {filepath} && " \
        "chmod {mode} {filepath} ".format(
            filepath=filepath, user=user, user_group=user_group, mode=mode,
            missing_owner_code=missing_owner_code)

    with settings(warn_only=True):
        result = sudo(command)
        if result.return_code == missing_owner_code:
            abort("User %s does not exist. Make sure the Presto server RPM "
                  "is installed and try again" % user)
        elif result.failed:
            abort("Failed to securely create file %s" % (filepath))


def deploy_node_properties(content, remote_dir):
    _LOGGER.info("Deploying node.properties configuration")
    name = "node.properties"
    node_file_path = (os.path.join(remote_dir, name))
    if not exists(node_file_path, use_sudo=True):
        secure_create_file(node_file_path, PRESTO_STANDALONE_USER_GROUP, mode=600)
    else:
        sudo('chown %(owner)s %(filepath)s && chmod %(mode)s %(filepath)s'
             % {'owner': PRESTO_STANDALONE_USER_GROUP, 'mode': 600, 'filepath': node_file_path})
    node_id_command = (
        "if ! ( grep -q -s 'node.id' " + node_file_path + " ); then "
        "uuid=$(uuidgen); "
        "echo node.id=$uuid >> " + node_file_path + ";"
        "fi; "
        "sed -i '/node.id/!d' " + node_file_path + "; "
        )
    sudo(node_id_command)
    files.append(os.path.join(remote_dir, name), content, True, shell=True)


def write_to_remote_file(text, filepath, owner, mode=600):
    secure_create_file(filepath, owner, mode)
    command = "echo '{text}' > {filepath}".format(
        text=escape_single_quotes(text), filepath=filepath)
    sudo(command)


def escape_single_quotes(text):
    # replace a single quote with a (closing) single quote followed by
    # an escaped quote followed by an (opening) single quote
    return text.replace(r"'", r"'\''")



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Logic at the application level for logging and exception handling"""

import functools
import logging
import logging.config
import os
import sys
import traceback

import __main__ as main

from trinoadmin import __version__
from trinoadmin.util import constants
from trinoadmin.util import filesystem
from trinoadmin.util.exception import ExceptionWithCause
from trinoadmin.util.local_config_util import get_log_directory

# Normally this would use the logger for __name__, however, this is
# effectively the "root" logger for the application.  If this code
# were running directly in the executable script __name__ would be
# set to '__main__', so we emulate that same behavior here.  This should
# resolve to the same logger that will be used by the entry point script.
logger = logging.getLogger('__main__')


class Application(object):
    """
    A generic application entry point.  Provides logging and exception handling
    features.  This class is expected to be used as a base class for various
    applications.

    Parameters:
        name - human readable name for the application
        version - the version of the application, as a string
        log_file_path - optional name of the log file including whatever
        extension you may want to use.  For example, 'foo.log' would create
        a file called 'foo.log' in the default presto-admin logging directory
        tree.
    """

    def __init__(self, name, version=None, log_file_path=None):
        self.name = str(name)
        self.__log_file_path = log_file_path or (self.name + '.log')
        if not os.path.isabs(self.__log_file_path):
            self.__log_file_path = os.path.join(
                get_log_directory(),
                self.__log_file_path
            )

        self.version = version or __version__

    def __enter__(self):
        self.__configure_logging()
        return self

    def __configure_logging(self):
        try:
            for maybe_file_path in self.__logging_configuration_file_paths():
                if not os.path.exists(maybe_file_path):
                    continue
                else:
                    config_file_path = maybe_file_path

                filesystem.ensure_parent_directories_exist(
                    self.__log_file_path
                )
                logging.config.fileConfig(
                    config_file_path,
                    defaults={'log_file_path': self.__log_file_path},
                    disable_existing_loggers=False
                )

                self.__log_application_start()
                logger.debug(
                    'Loaded logging configuration from %s',
                    config_file_path
                )
                break
        except Exception as e:
            sys.stderr.write(
                'Please run %s with sudo.\n' % self.name
            )
            sys.stderr.flush()
            sys.exit(str(e))

    def __logging_configuration_file_paths(self):
        # Current working directory
        yield constants.LOGGING_CONFIG_FILE_NAME
        # Application specific
        yield (self.__log_file_path + '.ini')
        yield (self.__main_module_path() + '.ini')
        # Global locations
        for dir_path in constants.LOGGING_CONFIG_FILE_DIRECTORIES:
            yield os.path.join(dir_path, constants.LOGGING_CONFIG_FILE_NAME)

    def __main_module_path(self):
        return os.path.abspath(main.__file__)

    def __log_application_start(self):
        LOG_SEPARATOR = '**************************************************'

        logger.debug(LOG_SEPARATOR)
        logger.debug(
            'Starting {name} {version}'.format(
                name=self.name,
                version=self.version
            )
        )
        logger.debug(LOG_SEPARATOR)
        logger.debug('raw arguments = {0}'.format(sys.argv))

    def __exit__(self, exc_type, exception, trace):
        self.exc_type = exc_type
        self.exception = exception
        self.trace = trace

        try:
            if exc_type is None:
                self.__handle_no_exception()
            elif exc_type == SystemExit:
                self.__handle_system_exit()
            else:
                self._handle_error()
                sys.exit(1)
        finally:
            self._exit_cleanup_hook()

    def _exit_cleanup_hook(self):
        logging.shutdown()

    def __handle_no_exception(self):
        logger.debug('Exiting normally')

    def __handle_system_exit(self):
        # Unfortunately a SystemExit can be raised with all kinds of
        # wonky values.  This code attempts to determine the actual
        # exit status.
        code = None
        try:
            # according to the docs a None value for this is equivalent
            # to a 0 value.
            if self.exception is None or self.exception.code is None:
                code = 0
            else:
                code = int(self.exception.code)
        except ValueError:
            code = 1
        except AttributeError:
            # In Python 2.6, the exceptions are passed as strings sometimes.
            # Thus exception.code gets an AttributeError.
            try:
                code = int(self.exception)
            except ValueError:
                code = 1
        except:
            logger.exception("Unknown exception: %s" % str(self.exception))

        if code is not None:
            if code is not 0:
                self._log_exception()
            logger.debug('Application exiting with status %d', code)
        else:
            self._log_exception()
        sys.exit(code)

    def _handle_error(self):
        self._log_exception()
        self.__display_error_message(str(self.exception))

    def __display_error_message(self, message):
        log_file_path = self.__get_root_log_file_path()
        error_message = ''
        if log_file_path:
            error_message += '  More detailed information can be found in '
            error_message += log_file_path
        print >> sys.stderr, message + error_message

    def __get_root_log_file_path(self):
        for handler in logging.root.handlers:
            if isinstance(handler, logging.FileHandler):
                return handler.baseFilename
        return None

    def _log_exception(self):
        formatted_stack_trace = ''.join(
            traceback.format_exception(
                self.exc_type,
                self.exception,
                self.trace
            ) + [ExceptionWithCause.get_cause_if_supported(self.exception)]
        )

        logger.error(
            'Handling uncaught exception: {t}, "{ex}"\n{tb}'.format(
                t=self.exc_type,
                ex=str(self.exception),
                tb=formatted_stack_trace
            )
        )


def entry_point(name, version=None, log_file_path=None,
                application_class=Application):
    """
    A decorator for application entry points.  The decorated function will
    be wrapped in an Application object and executed in that safe environment.
    Note that decorating a function with this decorator will not actually
    cause it to be invoked.  You must explicitly call the function in the
    script.

    Parameters:
        name - human readable name for the application
        version - the version of the application, as a string
        log_file_path - optional name of the log file including whatever
        extension you may want to use.  For example, 'foo.log' would create
        a file called 'foo.log' in the default prestoadmin logging directory
        tree.
        application_class - Type of application to run. The default is
        Application but there can be subclasses of that class.
    """

    def application_decorator(method):
        @functools.wraps(method)
        def wrapped_application(*args, **kwargs):
            with application_class(
                    name,
                    version=version,
                    log_file_path=log_file_path
            ):
                return method(*args, **kwargs)

        return wrapped_application

    return application_decorator



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Stuff to handle version ranges.
"""

import re

TD_VERSION = re.compile(r'^\d+t$')


def split_version(version_string):
    # We split on '.' and '-' because ancient tagged versions had the tag
    # delimited by a '-'
    return re.split('\.|-', version_string.strip())


def get_int_or_t(x):
    try:
        return int(x)
    except ValueError as e:
        if x is 't':
            return x
        if x[-1] is 't':
            int(x[:-1])
            return x
        raise e


def is_int_or_t(x):
    try:
        get_int_or_t(x)
        return True
    except ValueError:
        return False


def strip_tag(version):
    """
    Strip any parts of the version that are not numeric components or t's
    We leave the 't' on numeric components if it's present.
    ['1', '2', 'THREE'] -> (1, 2)
    ['1', 'TWO', '3'] -> (1, 3)
    ['0', '115t', 'SNAPSHOT'] -> (0, '115t')
    ['ZERO', '123t'] -> (123t)
    ['0', '148', 't'] => (0, 148, 't')
    ['0', '148', 't', 0, 1] => (0, 148, 't', 0, 1)
    ['0', '148', 't', 0, 1, 'SNAPSHOT'] => (0, 148, 't', 0, 1)
    ['0', '162', 'SNAPSHOT', 't', 'SNAPSHOT'] => (0, 162, 't')

    This checks the components of the version from least to most significant.

    :param version: something that can be sliced
    :return: a tuple containing only integer components or the letter t
    """

    result = list(version[:])
    result = [get_int_or_t(x) for x in result if is_int_or_t(x)]
    return tuple(result)


class VersionRange(object):
    """
    Represents a range of version numbers [min_version, max_version).
    The interval is right-open so that you can construct a numerically
    continuous list of versions like so:
    l = [VersionRange((0, 0), (0, 5)), VersionRange((0, 5), (1, 0))]
    and for all versions v where 0.0 <= v < 1.0 is contained in exactly one
    VersionRange in l.

    Continuity between version ranges can be checked using is_continuous.

    VersionRanges understand how to check if a Teradata version is contained
    in a VersionRange, but do no special handling to accomodate Teradata
    versions in their internal min_version and max_version members. I.e.,
    creating a VersionRange with a Teradata version will work, but __contains__
    will not work correctly. We don't currently need this, and hope not to.

    Note that the right-open interval representation of a version range does
    NOT allow the creation of a VersionRange that contains exactly one version.

    Note that empty intervals cannot be constructed as the serve no useful
    purpose. Specifically, we assert that min_version < max_version in the
    constructor.
    """

    def __init__(self, min_version, max_version, versioned_thing=None):
        # not pythonic, but bare ints screw things up.
        assert isinstance(min_version, tuple)
        assert isinstance(max_version, tuple)
        l = max(len(min_version), len(max_version))
        min_pad = VersionRange.pad_tuple(min_version, l, 0)
        max_pad = VersionRange.pad_tuple(max_version, l, 0)
        assert min_pad < max_pad
        self.min_version = min_version
        self.max_version = max_version
        self.versioned_thing = versioned_thing

    def __str__(self):
        return '[%s, %s) -> %s' % (
            '.'.join([str(c) for c in self.min_version]),
            '.'.join([str(c) for c in self.max_version]),
            self.versioned_thing)

    @staticmethod
    def strip_td_suffix(version):
        new_version = ()
        for component in version:
            if TD_VERSION.match(str(component)):
                new_last = component[:-1]
                new_version += (int(new_last),)
            elif component is not 't':
                new_version += (int(component),)

        return new_version

    @staticmethod
    def pad_tuple(tup, length, pad):
        assert len(tup) <= length
        result = list(tup)
        while len(result) < length:
            result.append(pad)
        return tuple(result)

    def zero_pad(self, other):
        """
        Pad out min_version, max_version, and other with zeroes to the length
        of the longest of the three. This allows subsequent comparisons to work
        as expected when tuples are of unequal length.
        Returns a tuple of tuples padded out to the same length
        """
        l = max(len(self.min_version), len(self.max_version), len(other))
        return (self.pad_tuple(self.min_version, l, 0),
                self.pad_tuple(self.max_version, l, 0),
                self.pad_tuple(other, l, 0))

    def __contains__(self, other):
        other = self.strip_td_suffix(other)
        other = tuple([int(component) for component in other])

        min_pad, max_pad, o_pad = self.zero_pad(other)
        return min_pad <= o_pad and o_pad < max_pad

    def is_continuous(self, next):
        min_pad, max_pad, next_min_pad = self.zero_pad(next.min_version)
        return max_pad == next_min_pad


class VersionRangeList(object):
    """
    A VersionRangeList is a list of continuous, non-overlapping VersionRanges.
    This is guaranteed by calling VersionRange.is_continuous on all pairs of
    VersionRanges vr[i], vr[i + 1] in the list, which ensures that the list is
    both sorted in order of ascending version and that the interval
    [vr[0].min_version, vr[n].max_version) has no discontinuities.
    """

    def __init__(self, *range_list):
        if len(range_list) >= 2:
            for i in range(0, len(range_list) - 1):
                assert range_list[i].is_continuous(range_list[i + 1])

        self.range_list = range_list

    def __str__(self):
        return '\n'.join([str(vr) for vr in self.range_list])

    def for_version(self, version):
        for range in self.range_list:
            if version in range:
                return range.versioned_thing
        raise KeyError(version)



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for managing presto/YARN integration.
"""

import os.path

from fabric.api import env, task, abort
from fabric.context_managers import shell_env
from fabric.operations import put, sudo, local

from trinoadmin.yarn_slider.config import SliderConfig, \
    DIR, SLIDER_USER, APPNAME, JAVA_HOME, HADOOP_CONF, SLIDER_MASTER, \
    PRESTO_PACKAGE, SLIDER_CONFIG_DIR
from trinoadmin.util.base_config import requires_config

from trinoadmin.util.fabricapi import task_by_rolename

__all__ = ['install', 'uninstall']


SLIDER_PKG_DEFAULT_FILES = ['appConfig-default.json', 'resources-default.json']


def get_slider_bin(conf):
    return os.path.join(conf[DIR], 'bin', 'slider')


def run_slider(slider_command, conf):
    with shell_env(JAVA_HOME=conf[JAVA_HOME],
                   HADOOP_CONF_DIR=conf[HADOOP_CONF]):
        return sudo(slider_command, user=conf[SLIDER_USER])


@task
@requires_config(SliderConfig)
@task_by_rolename(SLIDER_MASTER)
def install(presto_yarn_package):
    """
    Install the presto-yarn package on the cluster using Apache Slider. The
    presto-yarn package takes the form of a zip file that conforms to Slider's
    packaging requirements. After installing the presto-yarn package the presto
    application is registered with Slider.

    Before Slider can install the presto-yarn package, the slider user's hdfs
    home directory needs to be created. This needs to be done by a user that
    has write access to the hdfs /user directory, typically the user hdfs or a
    member of the superuser group.

    The name of the presto application is arbitrary and set in the slider
    configuration file. The default is PRESTO

    :param presto_yarn_package: The zip file containing the presto-yarn
                                package as structured for Slider.
    """
    conf = env.conf
    package_filename = os.path.basename(presto_yarn_package)
    package_file = os.path.join('/tmp', package_filename)

    result = put(presto_yarn_package, package_file)
    if result.failed:
        abort('Failed to send slider application package to %s on host %s' %
              (package_file, env.host))

    package_install_command = \
        '%s package --install --package %s --name %s' % \
        (get_slider_bin(conf), package_file, conf[APPNAME])

    try:
        run_slider(package_install_command, conf)

        conf[PRESTO_PACKAGE] = package_filename
        conf.store_conf()

        local('unzip %s %s -d %s' %
              (presto_yarn_package, ' '.join(SLIDER_PKG_DEFAULT_FILES),
               SLIDER_CONFIG_DIR))
    finally:
        sudo('rm -f %s' % (package_file))


@task
@requires_config(SliderConfig)
@task_by_rolename(SLIDER_MASTER)
def uninstall():
    """
    Uninstall unregisters the presto application with slider and removes the
    installed package.
    """
    conf = env.conf
    package_delete_command = '%s package --delete --name %s' % \
                             (get_slider_bin(conf), conf[APPNAME])
    run_slider(package_delete_command, conf)

    try:
        del conf[PRESTO_PACKAGE]
        conf.store_conf()
    except KeyError:
        pass

    local('rm %s' % (' '.join([os.path.join(SLIDER_CONFIG_DIR, f)
                     for f in SLIDER_PKG_DEFAULT_FILES])))



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import os
import time
from StringIO import StringIO

from fabric.context_managers import settings, hide
from fabric.operations import get, run
from fabric.state import env
from fabric.utils import error

from trinoadmin.config import get_conf_from_properties_data
from trinoadmin.util.constants import REMOTE_CONF_DIR, CONFIG_PROPERTIES, NODE_PROPERTIES

HTTP_ENABLED_KEY = 'http-server.http.enabled'
HTTPS_ENABLED_KEY = 'http-server.https.enabled'
HTTP_PORT_KEY = 'http-server.http.port'
HTTPS_PORT_KEY = 'http-server.https.port'
CLIENT_KEYSTORE_PATH_KEY = 'internal-communication.https.keystore.path'
CLIENT_KEYSTORE_PASSWORD_KEY = 'internal-communication.https.keystore.key'
INTERNAL_COMMUNICATION_SECRET = 'internal-communication.shared-secret'
NODE_ENVIRONMENT = 'node.environment'

_LOGGER = logging.getLogger(__name__)
# properties file literals
PROPERTIES_TRUE = 'true'
PROPERTIES_FALSE = 'false'


class PrestoConfig:
    # Defaults from Presto
    default_config = {
        HTTP_ENABLED_KEY: PROPERTIES_TRUE,
        HTTPS_ENABLED_KEY: PROPERTIES_FALSE,
        HTTP_PORT_KEY: '8080',
        HTTPS_PORT_KEY: '8443',
        CLIENT_KEYSTORE_PATH_KEY: None,
        CLIENT_KEYSTORE_PASSWORD_KEY: None,
        INTERNAL_COMMUNICATION_SECRET: None
    }

    default_node_config = {
        NODE_ENVIRONMENT: 'trino'
    }

    def __init__(self, config_properties, node_config_properties, config_path, config_host):
        self.config_path = config_path
        self.config_host = config_host
        if not node_config_properties:
            self.node_config_properties = self.default_node_config
        else:
            self.node_config_properties = node_config_properties
        if not config_properties:
            self.config_properties = self.default_config
        else:
            self.config_properties = config_properties

    @staticmethod
    def from_file(config, node_config, config_path=None, config_host=None):
        presto_config_dict = get_conf_from_properties_data(config)
        presto_node_config_dict = get_conf_from_properties_data(node_config)
        return PrestoConfig(presto_config_dict, presto_node_config_dict, config_path, config_host)

    @staticmethod
    def coordinator_config():
        config_path = os.path.join(REMOTE_CONF_DIR, CONFIG_PROPERTIES)
        config_host = env.roledefs['coordinator'][0]
        try:
            return PrestoConfig.from_file(
                PrestoConfig.read_file(config_host, config_path),
                PrestoConfig.read_file(config_host, os.path.join(REMOTE_CONF_DIR, NODE_PROPERTIES)),
                config_path,
                config_host)
        except:
            _LOGGER.info('Could not find Presto config.')
            return PrestoConfig(None, None, config_path, config_host)

    @staticmethod
    def read_file(config_host, config_path):
        data = StringIO()
        with settings(host_string='%s@%s' % (env.user, config_host)):
            with hide('stderr', 'stdout'):
                temp_dir = run('mktemp -d /tmp/trinoadmin.XXXXXXXXXXXXXX')
            try:
                get(config_path, data, use_sudo=True, temp_dir=temp_dir)
            finally:
                run('rm -rf %s' % temp_dir)
        data.seek(0)
        return data

    def _lookup(self, key):
        result = self.config_properties.get(key, self.default_config[key])
        if not result:
            error(
                    "Key %s is not configured in coordinator configuration"
                    "%s on host %s and has no default" %
                    (key, self.config_host, self.config_path))
        return result

    def _lookup_node_config(self, key):
        result = self.node_config_properties.get(key, self.default_node_config[key])
        if not result:
            error(
                "Key %s is not configured in coordinator node configuration"
                "%s on host %s and has no default" %
                (key, self.config_host, os.path.join(REMOTE_CONF_DIR, NODE_PROPERTIES)))
        return result

    def use_https(self):
        http_enabled = self._lookup(HTTP_ENABLED_KEY) == PROPERTIES_TRUE
        https_enabled = self._lookup(HTTPS_ENABLED_KEY) == PROPERTIES_TRUE

        return https_enabled and not http_enabled

    def get_client_keystore_path(self):
        return self._lookup(CLIENT_KEYSTORE_PATH_KEY)

    def get_client_keystore_password(self):
        return self._lookup(CLIENT_KEYSTORE_PASSWORD_KEY)

    def get_https_port(self):
        return int(self._lookup(HTTPS_PORT_KEY))

    def get_http_port(self):
        return int(self._lookup(HTTP_PORT_KEY))

    def get_internal_communication_secret(self):
        return self.config_properties.get(
            INTERNAL_COMMUNICATION_SECRET,
            self.default_config[INTERNAL_COMMUNICATION_SECRET])

    def get_node_environment(self):
        return self._lookup_node_config(NODE_ENVIRONMENT)



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for validating configuration information supplied by the user.
"""
import re
import socket

from fabric.context_managers import settings
from fabric.operations import run, sudo

from trinoadmin.util.exception import ConfigurationError


def validate_username(username):
    if not isinstance(username, basestring):
        raise ConfigurationError('Username must be of type string.')
    return username


def validate_port(port):
    try:
        port_int = int(port)
    except TypeError:
        raise ConfigurationError('Port must be of type string, but '
                                 'found ' + str(type(port)) + '.')
    except ValueError:
        raise ConfigurationError('Invalid port number ' + port +
                                 ': port must be a number between 1 and 65535')
    if not port_int > 0 or not port_int < 65535:
        raise ConfigurationError('Invalid port number ' + port +
                                 ': port must be a number between 1 and 65535')
    return port_int


def validate_host(host):
    try:
        socket.inet_pton(socket.AF_INET, host)
        return host
    except TypeError:
        raise ConfigurationError('Host must be of type string.  Found ' +
                                 str(type(host)) + '.')
    except socket.error:
        pass

    try:
        socket.inet_pton(socket.AF_INET6, host)
        return host
    except socket.error:
        pass

    if not is_valid_hostname(host):
        raise ConfigurationError(repr(host) + ' is not a valid '
                                 'ip address or host name.')
    return host


def is_valid_hostname(hostname):
    valid_name = '^(([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]*[a-zA-Z0-9])\.)*' \
                 '([A-Za-z0-9]|[A-Za-z0-9][A-Za-z0-9\-]*[A-Za-z0-9])$'
    return re.match(valid_name, hostname)


def validate_can_connect(user, host, port):
    with settings(host_string='%s@%s:%d' % (user, host, port), user=user):
        return run('exit 0').succeeded


def validate_can_sudo(sudo_user, conn_user, host, port):
    with settings(host_string='%s@%s:%d' % (conn_user, host, port),
                  warn_only=True):
        return sudo('exit 0', user=sudo_user).succeeded



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
module for tasks relating to presto plugins
"""
import logging
from fabric.decorators import task
from fabric.operations import sudo, put
import os
from fabric.api import env
from trinoadmin.standalone.config import StandaloneConfig
from trinoadmin.util.base_config import requires_config
from trinoadmin.util.constants import REMOTE_PLUGIN_DIR

__all__ = ['add_jar']
_LOGGER = logging.getLogger(__name__)


def write(local_path, remote_dir):
    sudo("mkdir -p " + remote_dir)
    put(local_path, remote_dir, use_sudo=True)


@task
@requires_config(StandaloneConfig)
def add_jar(local_path, plugin_name, plugin_dir=REMOTE_PLUGIN_DIR):
    """
    Deploy jar for the specified plugin to the plugin directory.

    Parameters:
        local_path - Local path to the jar to be deployed
        plugin_name - Name of the plugin subdirectory to deploy jars to
        plugin_dir - (Optional) The plugin directory.  If no directory is
                     given, '/usr/lib/presto/plugin' is used by default.
    """
    _LOGGER.info('deploying jars on %s' % env.host)
    write(local_path, os.path.join(plugin_dir, plugin_name))



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for the presto coordinator's configuration.
Loads and validates the coordinator.json file and creates the files needed
to deploy on the presto cluster
"""
from abc import abstractmethod, ABCMeta
import logging
import os

import config
import trino_conf
from trinoadmin.trino_conf import get_presto_conf

_LOGGER = logging.getLogger(__name__)


class Node():
    __metaclass__ = ABCMeta

    def __init__(self):
        pass

    def get_conf(self):
        conf = get_presto_conf(self._get_conf_dir())
        for name in trino_conf.REQUIRED_FILES:
            if name not in conf:
                _LOGGER.debug('%s configuration for %s not found.  '
                              'Default configuration will be deployed',
                              type(self).__name__, name)
                conf_value = self.default_config(name)
                conf[name] = conf_value
                file_path = os.path.join(self._get_conf_dir(), name)
                config.write_conf_to_file(conf_value, file_path)

        self.validate(conf)
        return conf

    @abstractmethod
    def _get_conf_dir(self):
        pass

    @abstractmethod
    def default_config(self, filename):
        pass

    @staticmethod
    @abstractmethod
    def validate(conf):
        pass

    def build_all_defaults(self):
        conf = {}
        for name in trino_conf.REQUIRED_FILES:
            conf[name] = self.default_config(name)
        return conf



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Filesystem tools."""

import errno
import logging
import os


logger = logging.getLogger(__name__)


def ensure_parent_directories_exist(path):
    try:
        os.makedirs(os.path.dirname(path))
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise e


def ensure_directory_exists(path):
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise e


def write_to_file_if_not_exists(content, path):
    flags = os.O_CREAT | os.O_EXCL | os.O_WRONLY

    try:
        os.makedirs(os.path.dirname(path))
    except OSError as e:
        if e.errno == errno.EEXIST:
            pass
        else:
            raise

    try:
        file_handle = os.open(path, flags)
    except OSError as e:
        if e.errno == errno.EEXIST:
            pass
        else:
            raise
    else:
        with os.fdopen(file_handle, 'w') as f:
            f.write(content)



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for various configuration management tasks using presto-admin
"""
import logging
import os
from StringIO import StringIO
from contextlib import closing

from fabric.contrib import files
from fabric.decorators import task, serial
from fabric.operations import get, sudo
from fabric.state import env
from fabric.utils import abort, warn

import trinoadmin.deploy
from trinoadmin.standalone.config import StandaloneConfig
from trinoadmin.util import constants
from trinoadmin.util.base_config import requires_config
from trinoadmin.util.constants import CONFIG_PROPERTIES, LOG_PROPERTIES, \
    JVM_CONFIG, NODE_PROPERTIES

__all__ = ['show']

ALL_CONFIG = [CONFIG_PROPERTIES, LOG_PROPERTIES, JVM_CONFIG, NODE_PROPERTIES]

_LOGGER = logging.getLogger(__name__)

__all__ = ['deploy', 'show']


@task
@requires_config(StandaloneConfig)
def deploy(rolename=None):
    """
    Deploy configuration on the remote hosts.

    Possible arguments are -
        coordinator - Deploy the coordinator configuration to the coordinator
        node
        workers - Deploy workers configuration to the worker nodes. This will
        not deploy configuration for a coordinator that is also a worker

    If no rolename is specified, then configuration for all roles will be
    deployed.  If there is no presto configuration file found in the
    configuration directory, default files will be deployed

    Parameters:
        rolename - [coordinator|workers]
    """
    if rolename is None:
        _LOGGER.info("Running configuration deploy")
        trinoadmin.deploy.coordinator()
        trinoadmin.deploy.workers()
    else:
        if rolename.lower() == 'coordinator':
            trinoadmin.deploy.coordinator()
        elif rolename.lower() == 'workers':
            trinoadmin.deploy.workers()
        else:
            abort("Invalid Argument. Possible values: coordinator, workers")


"""
gather/deploy_config_directory are used for server upgrade when we want to
preserve any existing configuration files across the upgrade exactly as they
were before the upgrade.

In order to preserve not just the data, but also the metadata, we tar up the
contents of /etc/presto to a temporary tar archive under /tmp. After the
upgrade, we untar it into /etc/presto and delete the archive.
"""


def gather_config_directory():
    """
    For the benefit of the next person to hack this, a list of some things
    that didn't work:
    - passing combine_stderr=False to sudo. Dunno why, still got them
    combined in the output.
    - using a StringIO object cfg = StringIO() and passing stdout=cfg. Got
    the host information at the start of the line.
    - sucking the tar archive over the network into memory instead of
    writing it out to a temporary file on the remote host. Since fabric
    doesn't provide a stdin= kwarg, there's no way to send back a tar
    archive larger than we can fit in a single bash command (~2MB on a
    good day), meaning if /etc/presto contains any large files, we'd end
    up having to send the archive to a temp file anyway.
    """
    result = sudo(
        'tarfile=`mktemp /tmp/presto_config-XXXXXXX.tar`; '
        'tar -c -z -C %s -f "${tarfile}" . && echo "${tarfile}"' % (
            constants.REMOTE_CONF_DIR,))
    return result


def deploy_config_directory(tarfile):
    sudo('tar -C "%s" -x -v -f "%s" ; rm "%s"' %
         (constants.REMOTE_CONF_DIR, tarfile, tarfile))


def configuration_fetch(file_name, config_destination, should_warn=True):
    remote_file_path = os.path.join(constants.REMOTE_CONF_DIR, file_name)
    if not files.exists(remote_file_path):
        if should_warn:
            warn("No configuration file found for %s at %s"
                 % (env.host, remote_file_path))
        return None
    else:
        get(remote_file_path, config_destination, use_sudo=True)
        return remote_file_path


def configuration_show(file_name, should_warn=True):
    with closing(StringIO()) as file_content_buffer:
        file_path = configuration_fetch(file_name, file_content_buffer,
                                        should_warn)
        if file_path is None:
            return
        config_values = file_content_buffer.getvalue()
        file_content_buffer.close()
        print ("\n%s: Configuration file at %s:" % (env.host, file_path))
        print config_values


@task
@requires_config(StandaloneConfig)
@serial
def show(config_type=None):
    """
    Print to the user the contents of the configuration files deployed

    If no config_type is specified, then all four configurations will be
    printed.  No warning will be printed for a missing log.properties since
    it is not a required configuration file.

    Parameters:
        config_type: [node|jvm|config|log]
    """
    file_name = ''
    if config_type is None:
        configuration_show(NODE_PROPERTIES)
        configuration_show(JVM_CONFIG)
        configuration_show(CONFIG_PROPERTIES)
        configuration_show(LOG_PROPERTIES, should_warn=False)
    else:
        if config_type.lower() == 'node':
            file_name = NODE_PROPERTIES
        elif config_type.lower() == 'jvm':
            file_name = JVM_CONFIG
        elif config_type.lower() == 'config':
            file_name = CONFIG_PROPERTIES
        elif config_type.lower() == 'log':
            file_name = LOG_PROPERTIES
        else:
            abort("Invalid Argument. Possible values: node, jvm, config, log")

        configuration_show(file_name)



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for the presto worker`'s configuration.
Loads and validates the workers.json file and creates the files needed
to deploy on the presto cluster
"""

import copy
import logging
import urlparse

from fabric.api import env

import trinoadmin.util.fabricapi as util
from trinoadmin.node import Node
from trinoadmin.trino_conf import validate_presto_conf
from trinoadmin.util import constants
from trinoadmin.util.exception import ConfigurationError
from trinoadmin.util.local_config_util import get_workers_directory

_LOGGER = logging.getLogger(__name__)


class Worker(Node):
    DEFAULT_PROPERTIES = {'node.properties':
                              {'node.environment': constants.BRAND,
                               'node.data-dir': '/var/lib/{}/data'.format(constants.BRAND),
                               'node.launcher-log-file':
                                   '/var/log/{}/launcher.log'.format(constants.BRAND),
                               'node.server-log-file':
                                   '/var/log/{}/server.log'.format(constants.BRAND),
                               'catalog.config-dir': '/etc/{}/catalog'.format(constants.BRAND),
                               'plugin.dir': '/usr/lib/{}/lib/plugin'.format(constants.BRAND)},
                          'jvm.config': ['-server',
                                         '-Xmx16G',
                                         '-XX:-UseBiasedLocking',
                                         '-XX:+UseG1GC',
                                         '-XX:G1HeapRegionSize=32M',
                                         '-XX:+ExplicitGCInvokesConcurrent',
                                         '-XX:+ExitOnOutOfMemoryError',
                                         '-XX:+UseGCOverheadLimit',
                                         '-XX:+HeapDumpOnOutOfMemoryError',
                                         '-XX:ReservedCodeCacheSize=512M',
                                         '-Djdk.attach.allowAttachSelf=true',
                                         '-Djdk.nio.maxCachedBufferSize=2000000',
                                         '-DHADOOP_USER_NAME=hive',  # not Presto default
                                         ],
                          'config.properties': {'coordinator': 'false',
                                                'http-server.http.port':
                                                    '8080',
                                                'query.max-memory': '50GB',
                                                'query.max-memory-per-node':
                                                    '8GB'}
                          }

    def _get_conf_dir(self):
        return get_workers_directory()

    def default_config(self, filename):
        try:
            conf = copy.deepcopy(self.DEFAULT_PROPERTIES[filename])
        except KeyError:
            raise ConfigurationError('Invalid configuration file name: %s' %
                                     filename)
        if filename == 'config.properties':
            coordinator = util.get_coordinator_role()[0]
            conf['discovery.uri'] = 'http://%s:8080' % coordinator
        return conf

    @staticmethod
    def is_localhost(hostname):
        return hostname in ['localhost', '127.0.0.1', '::1']

    @staticmethod
    def validate(conf):
        validate_presto_conf(conf)
        if 'coordinator' not in conf['config.properties']:
            raise ConfigurationError('Must specify coordinator=false in '
                                     'worker\'s config.properties')
        if conf['config.properties']['coordinator'] != 'false':
            raise ConfigurationError('Coordinator must be false in the '
                                     'worker\'s config.properties')
        uri = urlparse.urlparse(conf['config.properties']['discovery.uri'])
        if Worker.is_localhost(uri.hostname) and len(env.roledefs['all']) > 1:
            raise ConfigurationError(
                'discovery.uri should not be localhost in a '
                'multi-node cluster, but found ' + urlparse.urlunparse(uri) +
                '.  You may have encountered this error by '
                'choosing a coordinator that is localhost and a worker that '
                'is not.  The default discovery-uri is '
                'http://<coordinator>:8080')
        return conf



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Presto-Admin tool for deploying and managing Presto clusters"""

import os
import sys
import trinoadmin._version

from fabric.api import env

main_dir = os.path.dirname(os.path.abspath(os.path.dirname(__file__)))

import fabric_patches  # noqa

from trinoadmin.mode import get_mode, for_mode, MODE_STANDALONE, \
        MODE_SLIDER  # noqa
from trinoadmin.util.exception import ConfigFileNotFoundError, \
    ConfigurationError  # noqa

__version__ = trinoadmin._version.__version__

#
# Subcommands common to all modes. If anybody knows why fabric_patches is in
# the list, I'll make a note for the next person.
#
__all__ = ['fabric_patches']

cfg_mode = MODE_STANDALONE
try:
    cfg_mode = get_mode()
except ConfigFileNotFoundError as e:
    pass
except ConfigurationError as e:
    print >>sys.stderr, e.message


ADDITIONAL_TASK_MODULES = {
    MODE_SLIDER: [('yarn_slider.server', 'server'),
                  ('yarn_slider.slider', 'slider')],
    MODE_STANDALONE: ['topology',
                      ('configure_cmds', 'configuration'),
                      'server',
                      'catalog',
                      'package',
                      'collect',
                      'file',
                      'plugin']}


if cfg_mode is not None:
    atms = for_mode(cfg_mode, ADDITIONAL_TASK_MODULES)
    for atm in atms:
        try:
            module, subcommand_name = atm
        except ValueError:
            module = atm
            subcommand_name = atm

        __all__.append(subcommand_name)

        components = module.split('.')

        if len(components) == 1:
            # The simple case...
            # import <module> as <subcommand_name>
            globals()[subcommand_name] = __import__(module, globals())
        else:
            # The complicated case:
            # import foo.bar doesn't actually import foo.bar; it imports foo.
            # This is why, for example, you can't to the following:

            # >>> import os.path
            # >>> path.join('foo', 'bar', 'baz', 'zot')
            #
            # Doing the equivalent of import yarn_slider.slider as slider
            # results in the global slider variable being assigned to the
            # yarn_slider module, which is NOT what we want.
            # Instead, we need to recursively traverse the submodules until we
            # get to the one we're interested in.
            submodule = __import__(module, globals())
            for c in components[1:]:
                submodule = submodule.__dict__[c]
            globals()[subcommand_name] = submodule

env.roledefs = {
    'coordinator': [],
    'worker': [],
    'all': []
}



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Simple client to communicate with a Presto server.
"""
import base64
import datetime
import hashlib
import json
import logging
import os
import socket
import subprocess
import textwrap
import urlparse
from StringIO import StringIO
from httplib import HTTPConnection, HTTPException
from tempfile import mkstemp

import jks
import jwt
from fabric.operations import get
from fabric.state import env
from fabric.utils import error

from trinoadmin.util import constants
from trinoadmin.util.exception import InvalidArgumentError
from trinoadmin.util.httpscacertconnection import HTTPSCaCertConnection
from trinoadmin.util.local_config_util import get_coordinator_directory, get_topology_path
from trinoadmin.util.trino_config import PrestoConfig

_LOGGER = logging.getLogger(__name__)
URL_TIMEOUT_MS = 5000
NUM_ROWS = 1000
DATA_RESP = "data"
NEXT_URI_RESP = "nextUri"

CERTIFICATE_ALIAS = 'certificate_alias'


class TrinoClient:
    def __init__(self, server, user, coordinator_config=None):
        _LOGGER.info("Begin to initital TrinoClient class")
        # immutable stuff
        self.server = server
        self.user = user
        if coordinator_config is None:
            coordinator_config = PrestoConfig.coordinator_config()
        self.coordinator_config = coordinator_config
        self.port = TrinoClient._get_configured_port(self.coordinator_config)
        # mutable stuff
        self.ca_file_path = ""
        self.keystore_data = ""
        self.rows = []
        self.next_uri = ''
        self.response_from_server = {}
        self.cli = None
        # find the presto/trino client
        _LOGGER.debug("Try to find trino/presto command line")
        find_cli = "which " + constants.BRAND
        res = subprocess.check_output(find_cli, shell=True)
        if res:
            self.cli = res.decode().strip()
            _LOGGER.info("use command line {}".format(self.cli))
        else:
            _LOGGER.warn("Can not find presto/trino command line")

    @staticmethod
    def _remove_silently(path):
        try:
            os.remove(path)
        except:
            pass

    def close(self):
        TrinoClient._remove_silently(self.ca_file_path)

    def _clear_old_results(self):
        if self.rows:
            self.rows = []

        if self.next_uri:
            self.next_uri = ''

        if self.response_from_server:
            self.response_from_server = {}

    def run_sql(self, sql, schema="default", catalog="hive"):
        """
        Execute a query connecting to Presto server using passed parameters.

        Args:
            sql: SQL query to be executed
            schema: Presto schema to be used while executing query
                (default=default)
            catalog: Catalog to be used by the server

        Returns:
            list of rows or None if client was unable to connect to Presto
        """
        result = []
        if self.cli:
            cmd = "%s --server %s:%s --catalog %s --schema %s --output-format=CSV_UNQUOTED --execute \"%s\"" % \
                  (self.cli, self.server, self.port, catalog, schema, sql)
            output = subprocess.check_output(cmd, shell=True).strip().split('\n')

            for row in output:
                result.append(row.split(","))
            return result
        else:
            status = self._execute_query(sql, schema, catalog)
            if status:
                return self._get_rows()
            else:
                return None

    def _execute_query(self, sql, schema, catalog):
        if not sql:
            raise InvalidArgumentError("SQL query missing")

        if not self.server:
            raise InvalidArgumentError("Server IP missing")

        if not self.user:
            raise InvalidArgumentError("Username missing")

        self._clear_old_results()

        if constants.BRAND == 'trino':
            headers = {"X-Trino-Catalog": catalog,
                       "X-Trino-Schema": schema,
                       "X-Trino-User": self.user,
                       "X-Trino-Source": "trino-admin"}
        else:
            headers = {"X-Presto-Catalog": catalog,
                       "X-Presto-Schema": schema,
                       "X-Presto-User": self.user,
                       "X-Presto-Source": "presto-admin"}
        answer = ''
        try:
            _LOGGER.info("Connecting to server at: " + self.server +
                         ":" + str(self.port) + " as user " + self.user +
                         " to execute query " + sql)
            conn = self._get_connection()
            self._add_auth_headers(headers)
            conn.request("POST", "/v1/statement", sql, headers)
            response = conn.getresponse()

            if response.status != 200:
                conn.close()
                _LOGGER.error("Connection error: " +
                              str(response.status) + " " + response.reason)
                return False

            answer = response.read()
            conn.close()

            self.response_from_server = json.loads(answer)
            _LOGGER.info("Query executed successfully: %s" % (sql))
            return True
        except (HTTPException, socket.error) as e:
            _LOGGER.error("Error connecting to presto server at: " +
                          self.server + ":" + str(self.port) + ' ' + e.message)
            return False
        except ValueError as e:
            _LOGGER.error('Error connecting to Presto server: ' + e.message +
                          ' error from server: ' + answer)
            raise e

    def _get_response_from(self, uri):
        """
        Sends a GET request to the Presto server at the specified next_uri
        and updates the response

        Remove the scheme and host/port from the uri; the connection itself
        has that information.
        """
        parts = list(urlparse.urlsplit(uri))
        parts[0] = None
        parts[1] = None
        location = urlparse.urlunsplit(parts)
        conn = self._get_connection()
        headers = {"X-" + constants.BRAND.capitalize() + "-User": self.user}
        self._add_auth_headers(headers)
        conn.request("GET", location, headers=headers)
        response = conn.getresponse()

        if response.status != 200:
            conn.close()
            _LOGGER.error("Error making GET request to %s: %s %s" %
                          (uri, response.status, response.reason))
            return False

        answer = response.read()
        conn.close()

        self.response_from_server = json.loads(answer)
        _LOGGER.info("GET request successful for uri: " + uri)
        return True

    def _build_results_from_response(self):
        """
        Build result from the response

        The reponse_from_server may contain up to 3 uri's.
        1. link to fetch the next packet of data ('nextUri')
        2. TODO: information about the query execution ('infoUri')
        3. TODO: cancel the query ('partialCancelUri').
        """
        if NEXT_URI_RESP in self.response_from_server:
            self.next_uri = self.response_from_server[NEXT_URI_RESP]
        else:
            self.next_uri = ""

        if DATA_RESP in self.response_from_server:
            if self.rows:
                self.rows.extend(self.response_from_server[DATA_RESP])
            else:
                self.rows = self.response_from_server[DATA_RESP]

    def _get_rows(self, num_of_rows=NUM_ROWS):
        """
        Get the rows returned from the query.

        The client sends GET requests to the server using the 'nextUri'
        from the previous response until the servers response does not
        contain anymore 'nextUri's.  When there is no 'nextUri' the query is
        finished

        Note that this can only be called once and does not page through
        the results.

        Parameters:
            num_of_rows: to be retrieved. 1000 by default
        """
        if num_of_rows == 0:
            return []

        self._build_results_from_response()

        if not self._get_next_uri():
            return []

        while self._get_next_uri():
            if not self._get_response_from(self._get_next_uri()):
                return []
            if (len(self.rows) <= num_of_rows):
                self._build_results_from_response()
        return self.rows

    def _get_next_uri(self):
        return self.next_uri

    def _get_connection(self):
        if self.coordinator_config.use_https():
            return self._get_https_connection()
        else:
            return HTTPConnection(self.server, self.port, False, URL_TIMEOUT_MS)

    @staticmethod
    def _get_configured_port(coordinator_config):
        if coordinator_config.use_https():
            return coordinator_config.get_https_port()
        else:
            return coordinator_config.get_http_port()

    def _get_https_connection(self):
        ca_file_path = self._get_pem()
        result = HTTPSCaCertConnection(
            self.server, self.port, None, None, ca_file_path, False, URL_TIMEOUT_MS)
        return result

    def _fetch_keystore_data(self):
        if not self.keystore_data:
            remote_keystore_path = self.coordinator_config.get_client_keystore_path()
            keystore_data = StringIO()
            get(remote_keystore_path, keystore_data, use_sudo=True)
            keystore_data.seek(0)
            self.keystore_data = keystore_data.getvalue()
        return self.keystore_data

    def _pem_string(self, der_bytes, type):
        result = "-----BEGIN %s-----\n" % type
        result += "\r\n".join(
            textwrap.wrap(base64.b64encode(der_bytes).decode('ascii'), 64))
        result += "\n-----END %s-----\n" % type
        return result

    def _write_pem_file(self, directory, der_bytes_list, type):
        prefix = os.path.join(directory,
                              '%s-' % type.lower().replace(' ', '-'))
        fd, pem_path = mkstemp('.pem', prefix)
        # https://www.digicert.com/ssl-support/pem-ssl-creation.htm
        with open(pem_path, 'w') as pem_file:
            for der_bytes in der_bytes_list:
                pem_file.write(self._pem_string(der_bytes, type))
        os.close(fd)
        return pem_path

    def _get_pem(self):
        keystore_data = self._fetch_keystore_data()

        keystore = jks.KeyStore.loads(
            keystore_data,
            self.coordinator_config.get_client_keystore_password())

        if len(keystore.private_keys.items()) == 1:
            _, private_key = keystore.private_keys.items()[0]
        else:
            private_key = self._get_private_key(keystore)
        if not self.ca_file_path:
            """
            Each member of the cert chain is a tuple (cert_type, cert_data)
            We only need to write the data out to the .PEM file.

            This usage is shown in the example in the README.md on github:
            https://github.com/kurtbrose/pyjks
            """
            self.ca_file_path = self._write_pem_file(
                get_coordinator_directory(),
                [cert[1] for cert in private_key.cert_chain], 'CERTIFICATE')

        return self.ca_file_path

    def _get_private_key(self, keystore):
        all_keys = ", ".join(keystore.private_keys.keys())
        try:
            alias = env.conf[CERTIFICATE_ALIAS]
        except KeyError:
            error('Multiple keys found in %s. Set %s in %s. Available aliases are %s' %
                  (self.coordinator_config.get_client_keystore_path(),
                   CERTIFICATE_ALIAS, get_topology_path(), all_keys))

        try:
            return keystore.private_keys[alias]
        except KeyError:
            error('No alias %s found in %s. Available aliases are %s' %
                  (alias, self.coordinator_config.get_client_keystore_path(),
                   all_keys))

    def _add_auth_headers(self, headers):
        internal_secret = self.coordinator_config.get_internal_communication_secret()
        if not internal_secret:
            node_environment = self.coordinator_config.get_node_environment()
            _LOGGER.info('Using \'node.environment=%s\' for internal communication secret' % node_environment)
            internal_secret = node_environment
        auth_headers = self._create_auth_headers(internal_secret)
        headers.update(auth_headers)

    def _create_auth_headers(self, secret=None):
        if not secret:
            _LOGGER.warn('No internal secret found. '
                         'Fallback to using user %s for auth header X-Presto-User' % self.user)
            return {"X-Presto-User": self.user}

        return {'X-Presto-Internal-Bearer': '%s' % self.__generate_presto_token(secret.rstrip())}

    @staticmethod
    def __generate_presto_token(secret):
        key = hashlib.sha256(secret.encode()).digest()
        expire = datetime.datetime.utcnow() + datetime.timedelta(minutes=5)
        payload = {'sub': 'PrestoAdmin', 'exp': expire}
        return jwt.encode(payload, key, algorithm='HS256')



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
An option group for which you can hide the help text.
"""

import logging
from optparse import OptionGroup


_LOGGER = logging.getLogger(__name__)


class HiddenOptionGroup(OptionGroup):
    """
    Optparse allows you to suppress Options from the help text, but not
    groups. This class allows you to suppress the help of groups.
    """

    def __init__(self, parser, title, description=None, suppress_help=False):
        OptionGroup.__init__(self, parser, title, description)
        self.suppress_help = suppress_help

    def format_help(self, formatter):
        if not self.suppress_help:
            return OptionGroup.format_help(self, formatter)
        else:
            return ""



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for rpm package deploy and install using presto-admin
"""
import logging

from fabric.context_managers import settings, hide, shell_env
from fabric.decorators import task, runs_once
from fabric.operations import sudo, put, os, local
from fabric.state import env
from fabric.tasks import execute
from fabric.utils import abort

from trinoadmin.util import constants
from trinoadmin.standalone.config import StandaloneConfig
from trinoadmin.util.base_config import requires_config
from trinoadmin.util.fabricapi import get_host_list

_LOGGER = logging.getLogger(__name__)
__all__ = ['install', 'uninstall']


@task
@runs_once
@requires_config(StandaloneConfig)
def install(local_path):
    """
    Install the rpm package on the cluster

    Args:
        local_path: Absolute path to the rpm to be installed
        --nodeps (optional): Flag to indicate if rpm install
            should ignore checking package dependencies. Equivalent
            to adding --nodeps flag to rpm -i.
    """
    check_if_valid_rpm(local_path)
    return execute(deploy_install, local_path, hosts=get_host_list())


def check_if_valid_rpm(local_path):
    _LOGGER.info("Checking rpm checksum to see if it is corrupted")
    with settings(hide('warnings', 'stdout'), warn_only=True):
        result = local('rpm -K --nosignature ' + local_path, capture=True)
    if 'MD5 NOT OK' in result.stdout:
        abort("Corrupted RPM. Try downloading the RPM again.")
    elif result.stderr:
        abort(result.stderr)


def deploy_install(local_path):
    deploy_action(local_path, rpm_install)


def deploy_upgrade(local_path):
    deploy_action(local_path, rpm_upgrade)


def deploy_action(local_path, rpm_action):
    deploy(local_path)
    rpm_action(os.path.basename(local_path))


def deploy(local_path=None):
    if not os.path.isfile(local_path):
        abort('RPM file not found at %s.' % local_path)

    _LOGGER.info("Deploying rpm on %s..." % env.host)
    print("Deploying rpm on %s..." % env.host)
    sudo('mkdir -p ' + constants.REMOTE_PACKAGES_PATH)
    ret_list = put(local_path, constants.REMOTE_PACKAGES_PATH, use_sudo=True)
    if not ret_list.succeeded:
        _LOGGER.warn("Failure during put. Now using /tmp as temp dir...")
        ret_list = put(local_path, constants.REMOTE_PACKAGES_PATH,
                       use_sudo=True, temp_dir='/tmp')
    if ret_list.succeeded:
        print("Package deployed successfully on: " + env.host)


def _rpm_install(package_path):
    nodeps = _nodeps_rpm_option()

    if 'java_home' in env and env.java_home:
        with shell_env(JAVA_HOME='%s' % env.java_home):
            return sudo('rpm -i %s%s' % (nodeps, package_path))
    else:
        return sudo('rpm -i %s%s' % (nodeps, package_path))


def _nodeps_rpm_option():
    return '--nodeps ' if env.nodeps else ''


def rpm_install(rpm_name):
    _LOGGER.info("Installing the rpm")
    if _rpm_install(_rpm_path(rpm_name)).succeeded:
        print("Package installed successfully on: " + env.host)


def _rpm_path(rpm_filename):
    return os.path.join(constants.REMOTE_PACKAGES_PATH, rpm_filename)


def rpm_upgrade(rpm_name):
    _LOGGER.info("Upgrading the rpm")
    rpm_path = _rpm_path(rpm_name)
    package_name = sudo('rpm -qp --queryformat \'%%{NAME}\' %s' % rpm_path,
                        quiet=True)

    if not package_name.succeeded:
        abort("Corrupted RPM file: %s" % rpm_path)

    if _rpm_upgrade(rpm_path).succeeded:
        print("Package upgraded successfully on: " + env.host)


def _rpm_upgrade(package_name):
    return sudo('rpm -U %s%s' % (_nodeps_rpm_option(), package_name))


@task
@runs_once
@requires_config(StandaloneConfig)
def uninstall(rpm_name):
    """
    Uninstall the rpm package from the cluster

    Args:
        rpm_name: Name of the rpm to be uninstalled
        --nodeps (optional): Flag to indicate if rpm uninstall) should ignore checking package dependencies. Equivalent
            to adding --nodeps flag to rpm -e.
        --force (optional): Flag to indicate that rpm uninstall should not fail if package is not installed.
    """
    return execute(rpm_uninstall, rpm_name, hosts=get_host_list())


def rpm_uninstall(package_name):
    _LOGGER.info("Uninstalling the rpm")

    if not is_rpm_installed(package_name):
        if not env.force:
            abort('Package is not installed: ' + package_name)
    elif _rpm_uninstall(package_name).succeeded:
        print("Package uninstalled successfully on: " + env.host)


def is_rpm_installed(package_name):
    return sudo('rpm -qi %s' % package_name, quiet=True).succeeded


def _rpm_uninstall(package_name):
    return sudo('rpm -e %s%s' % (_nodeps_rpm_option(), package_name))



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for presto catalog configurations
"""
import errno
import logging

import fabric.utils
from fabric.api import task, env
from fabric.context_managers import hide
from fabric.contrib import files
from fabric.operations import sudo, os, get

from trinoadmin.deploy import secure_create_directory
from trinoadmin.standalone.config import StandaloneConfig, \
    PRESTO_STANDALONE_USER_GROUP
from trinoadmin.util import constants
from trinoadmin.util.base_config import requires_config
from trinoadmin.util.exception import ConfigFileNotFoundError, \
    ConfigurationError
from trinoadmin.util.fabricapi import put_secure
from trinoadmin.util.filesystem import ensure_directory_exists
from trinoadmin.util.local_config_util import get_catalog_directory

_LOGGER = logging.getLogger(__name__)

__all__ = ['add', 'remove']
COULD_NOT_REMOVE = 'Could not remove catalog'


# we deploy catalog files with 0600 permissions because they can contain passwords
# that should not be world readable
def deploy_files(filenames, local_dir, remote_dir, user_group, mode=0600):
    _LOGGER.info('Deploying configurations for ' + str(filenames))
    secure_create_directory(remote_dir, PRESTO_STANDALONE_USER_GROUP)
    for name in filenames:
        put_secure(user_group, mode, os.path.join(local_dir, name), remote_dir,
                   use_sudo=True)


def gather_catalogs(local_config_dir, allow_overwrite=False):
    local_catalog_dir = os.path.join(local_config_dir, env.host, 'catalog')
    if not allow_overwrite and os.path.exists(local_catalog_dir):
        fabric.utils.error("Refusing to overwrite %s. Use 'overwrite' "
                           "option to overwrite." % local_catalog_dir)
    ensure_directory_exists(local_catalog_dir)
    if files.exists(constants.REMOTE_CATALOG_DIR):
        return get(constants.REMOTE_CATALOG_DIR, local_catalog_dir, use_sudo=True)
    else:
        return []


def validate(filenames):
    for name in filenames:
        file_path = os.path.join(get_catalog_directory(), name)
        _LOGGER.info('Validating catalog configuration: ' + str(name))
        try:
            with open(file_path) as f:
                file_content = f.read()
            if 'connector.name' not in file_content:
                message = ('Catalog configuration %s does not contain '
                           'connector.name' % name)
                raise ConfigurationError(message)

        except IOError, e:
            fabric.utils.error(message='Error validating ' + file_path,
                               exception=e)
            return False

    return True


@task
@requires_config(StandaloneConfig)
def add(name=None):
    """
    Deploy configuration for a catalog onto a cluster.

    E.g.: 'presto-admin catalog add tpch'
    deploys a configuration file for the tpch connector.  The configuration is
    defined by tpch.properties in the local catalog directory, which defaults to
    ~/.trinoadmin/catalog.

    If no catalog name is specified, then  configurations for all catalogs
    in the catalog directory will be deployed

    Parameters:
        name - Name of the catalog to be added
    """
    catalog_dir = get_catalog_directory()
    if name:
        filename = name + '.properties'
        config_path = os.path.join(catalog_dir, filename)
        if not os.path.isfile(config_path):
            raise ConfigFileNotFoundError(
                config_path=config_path,
                message='Configuration for catalog ' + name + ' not found')
        filenames = [filename]
    elif not os.path.isdir(catalog_dir):
        message = ('Cannot add catalogs because directory %s does not exist'
                   % catalog_dir)
        raise ConfigFileNotFoundError(config_path=catalog_dir,
                                      message=message)
    else:
        try:
            filenames = os.listdir(catalog_dir)
        except OSError as e:
            fabric.utils.error(e.strerror)
            return
        if not filenames:
            fabric.utils.warn(
                'Directory %s is empty. No catalogs will be deployed' %
                catalog_dir)
            return

    if not validate(filenames):
        return
    filenames.sort()
    _LOGGER.info('Adding catalog configurations: ' + str(filenames))
    print('Deploying %s catalog configurations on: %s ' %
          (', '.join(filenames), env.host))

    deploy_files(filenames, catalog_dir,
                 constants.REMOTE_CATALOG_DIR, PRESTO_STANDALONE_USER_GROUP)


@task
@requires_config(StandaloneConfig)
def remove(name):
    """
    Remove a catalog from the cluster.

    Parameters:
        name - Name of the catalog to be removed
    """
    _LOGGER.info('[' + env.host + '] Removing catalog: ' + name)
    ret = remove_file(os.path.join(constants.REMOTE_CATALOG_DIR,
                                   name + '.properties'))
    if ret.succeeded:
        if COULD_NOT_REMOVE in ret:
            fabric.utils.error(ret)
        else:
            print('[%s] Catalog removed. Restart the server for the change '
                  'to take effect' % env.host)
    else:
        fabric.utils.error('Failed to remove catalog ' + name + '.\n\t' +
                           ret)

    local_path = os.path.join(get_catalog_directory(), name + '.properties')
    try:
        os.remove(local_path)
    except OSError as e:
        if e.errno == errno.ENOENT:
            pass
        else:
            raise


def remove_file(path):
    script = ('if [ -f %(path)s ] ; '
              'then rm %(path)s ; '
              'else echo "%(could_not_remove)s \'%(name)s\'. '
              'No such file \'%(path)s\'"; fi')

    with hide('stderr', 'stdout'):
        return sudo(script %
                    {'path': path,
                     'name': os.path.splitext(os.path.basename(path))[0],
                     'could_not_remove': COULD_NOT_REMOVE})
